import numpy as np
import jax.scipy.linalg as jla

import supergrad
from supergrad.quantum_system.kronobj import KronObj


def random_complex_matrix(dim):
    return rng.random((dim, dim)) + 1j * rng.random((dim, dim))


rng = np.random.default_rng(seed=42)
dims = [3, 5, 6, 4, 2]
mat0 = random_complex_matrix(dims[0])
mat1 = random_complex_matrix(dims[1])
mat3 = random_complex_matrix(dims[3])
kron1 = KronObj([mat0, mat1, mat3], dims, [0, 1, 3])
manual1 = supergrad.tensor(mat0, mat1, np.eye(dims[2]), mat3, np.eye(dims[4]))

mat2 = random_complex_matrix(dims[2])
mat4 = random_complex_matrix(dims[4])
kron2 = KronObj([mat2, mat4], dims, [2, 4])
manual2 = supergrad.tensor(np.eye(dims[0]), np.eye(dims[1]), mat2,
                           np.eye(dims[3]), mat4)


def test_kronecker_product():

    assert np.allclose(manual1, kron1.full())
    assert np.allclose(manual2, kron2.full())


def test_addition():

    assert np.allclose(manual1 + manual2, (kron1 + kron2).full())
    assert np.allclose(manual2 + manual1, (kron2 + kron1).full())
    assert np.allclose(manual1 - manual2, (kron1 - kron2).full())
    assert np.allclose(manual2 - manual1, (kron2 - kron1).full())


def test_dag():

    assert np.allclose(
        np.conj(manual1 + manual2).T, (kron1 + kron2).dag().full())


def test_scalar_multiplication():
    flt = rng.random()
    assert np.allclose(flt * (manual1 + manual2),
                       (flt * (kron1 + kron2)).full())
    assert np.allclose(flt * (manual1 + manual2),
                       ((kron1 + kron2) * flt).full())


def test_vector_multiplication():
    psi = rng.random(kron1.shape[0])[:, np.newaxis]
    assert np.allclose((manual1 + manual2) @ psi, ((kron1 + kron2) @ psi))


def test_matrix_exponential():
    res = jla.expm((manual1 + manual2))
    res0 = (kron1 + kron2).expm()
    assert np.allclose(res0, res)


def test_exponential_matrix_vecmul():
    psi = rng.random(kron2.shape[0])[:, np.newaxis]
    mixed_kron = kron1 + kron2
    assert np.allclose(kron1.expm(psi, 1), jla.expm(manual1) @ psi)
    assert np.allclose(kron2.expm(psi, 1), jla.expm(manual2) @ psi)
    assert np.allclose(mixed_kron.expm(psi, 1),
                       jla.expm(manual1 + manual2) @ psi)


def test_matmul_between_kronobj():
    assert np.allclose((kron1 @ kron2).full(), manual1 @ manual2)
    assert np.allclose(((kron1 + kron2) @ (kron1 + kron1)).full(),
                       (manual1 + manual2) @ (manual1 + manual1))



_release_version = '0.2.3'


def _get_version_for_build() -> str:
    """Determine the version at build time."""
    if _release_version is not None:
        return _release_version


def _get_version_string() -> str:
    # The build/source distribution could overwrite _release_version.
    if _release_version is not None:
        return _release_version


__version__ = _get_version_string()



"""
The code is from `jax._src.numpy.util `.
"""
from __future__ import annotations

from collections.abc import Callable, Sequence
import re
import textwrap
from typing import Any, NamedTuple, TypeVar

from jax._src import config

_T = TypeVar("_T")
_parameter_break = re.compile("\n(?=[A-Za-z_])")
_section_break = re.compile(r"\n(?=[^\n]{3,15}\n-{3,15})", re.MULTILINE)
_numpy_signature_re = re.compile(r'^([\w., ]+=)?\s*[\w\.]+\([\w\W]*?\)$',
                                 re.MULTILINE)
_versionadded = re.compile(r'^\s+\.\.\s+versionadded::', re.MULTILINE)
_docreference = re.compile(r':doc:`(.*?)\s*<.*?>`')


class ParsedDoc(NamedTuple):
    """
  docstr: full docstring
  signature: signature from docstring.
  summary: summary from docstring.
  front_matter: front matter before sections.
  sections: dictionary of section titles to section content.
  """
    docstr: str | None
    signature: str = ""
    summary: str = ""
    front_matter: str = ""
    sections: dict[str, str] = {}


def _parse_numpydoc(docstr: str | None) -> ParsedDoc:
    """Parse a standard numpy-style docstring.
  Args:
    docstr: the raw docstring from a function
  Returns:
    ParsedDoc: parsed version of the docstring
  """
    if docstr is None or not docstr.strip():
        return ParsedDoc(docstr)
    # Remove any :doc: directives in the docstring to avoid sphinx errors
    docstr = _docreference.sub(lambda match: f"{match.groups()[0]}", docstr)
    signature, body = "", docstr
    match = _numpy_signature_re.match(body)
    if match:
        signature = match.group()
        body = docstr[match.end():]
    firstline, _, body = body.partition('\n')
    body = textwrap.dedent(body.lstrip('\n'))
    match = _numpy_signature_re.match(body)
    if match:
        signature = match.group()
        body = body[match.end():]
    summary = firstline
    if not summary:
        summary, _, body = body.lstrip('\n').partition('\n')
        body = textwrap.dedent(body.lstrip('\n'))
    front_matter = ""
    body = "\n" + body
    section_list = _section_break.split(body)
    if not _section_break.match(section_list[0]):
        front_matter, *section_list = section_list
    sections = {section.split('\n', 1)[0]: section for section in section_list}
    return ParsedDoc(docstr=docstr,
                     signature=signature,
                     summary=summary,
                     front_matter=front_matter,
                     sections=sections)


def _parse_parameters(body: str) -> dict[str, str]:
    """Parse the Parameters section of a docstring."""
    title, underline, content = body.split('\n', 2)
    assert title == 'Parameters'
    assert underline and not underline.strip('-')
    parameters = _parameter_break.split(content)
    return {p.partition(' : ')[0].partition(', ')[0]: p for p in parameters}


def implements(
    original_fun: Callable[..., Any] | None,
    update_doc: bool = True,
    sections: Sequence[str] = ('Parameters', 'Returns', 'References'),
    module: str | None = None,
) -> Callable[[_T], _T]:
    """Decorator for SuperGrad functions which implement a specified NumPy function.
  This mainly contains logic to copy and modify the docstring of the original
  function. In particular, if `update_doc` is True, parameters listed in the
  original function that are not supported by the decorated function will
  be removed from the docstring. For this reason, it is important that parameter
  names match those in the original numpy function.
  Args:
    original_fun: The original function being implemented
    update_doc: whether to transform the numpy docstring to remove references of
      parameters that are supported by the numpy version but not the JAX version.
      If False, include the numpy docstring verbatim.
    sections: a list of sections to include in the docstring. The default is
      ["Parameters", "Returns", "References"]
    module: an optional string specifying the module from which the original function
      is imported. This is useful for objects such as ufuncs, where the module cannot
      be determined from the original function itself.
  """

    def decorator(wrapped_fun):
        wrapped_fun.__np_wrapped__ = original_fun
        # Allows this pattern: @implements(getattr(np, 'new_function', None))
        if original_fun is None:
            return wrapped_fun
        docstr = getattr(original_fun, "__doc__", None)
        name = getattr(original_fun, "__name__",
                       getattr(wrapped_fun, "__name__", str(wrapped_fun)))
        try:
            mod = module or original_fun.__module__
        except AttributeError:
            if config.enable_checks.value:
                raise ValueError(
                    f"function {original_fun} defines no __module__; pass module keyword to implements()."
                )
        else:
            name = f"{mod}.{name}"
        if docstr:
            try:
                parsed = _parse_numpydoc(docstr)
                if update_doc and 'Parameters' in parsed.sections:
                    code = getattr(
                        getattr(wrapped_fun, "__wrapped__", wrapped_fun),
                        "__code__", None)
                    # Remove unrecognized parameter descriptions.
                    parameters = _parse_parameters(
                        parsed.sections['Parameters'])
                    parameters = {
                        p: desc
                        for p, desc in parameters.items()
                        if (code is None or p in code.co_varnames)
                    }
                    if parameters:
                        parsed.sections['Parameters'] = (
                            "Parameters\n"
                            "----------\n" + "\n".join(
                                _versionadded.split(desc)[0].rstrip()
                                for p, desc in parameters.items()))
                    else:
                        del parsed.sections['Parameters']
                docstr = parsed.summary.strip() + "\n" if parsed.summary else ""
                docstr += f"\nSuperGrad-supported implementation of :func:`{name}`.\n"
                docstr += "\n*Original docstring below.*\n"
                # We remove signatures from the docstrings, because they redundant at best and
                # misleading at worst: e.g. JAX wrappers don't implement all ufunc keyword arguments.
                # if parsed.signature:
                #   docstr += "\n" + parsed.signature.strip() + "\n"
                if parsed.front_matter:
                    docstr += "\n" + parsed.front_matter.strip() + "\n"
                kept_sections = (
                    content.strip()
                    for section, content in parsed.sections.items()
                    if section in sections)
                if kept_sections:
                    docstr += "\n" + "\n\n".join(kept_sections) + "\n"
            except:
                if config.enable_checks.value:
                    raise
                docstr = original_fun.__doc__
        wrapped_fun.__doc__ = docstr
        for attr in ['__name__', '__qualname__']:
            try:
                value = getattr(original_fun, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapped_fun, attr, value)
        return wrapped_fun

    return decorator



from supergrad.scgraph.graph import SCGraph
from supergrad.scgraph.graph_mpc_fluxonium_1d import MPCFluxonium1D
from supergrad.scgraph.graph_mpc_fluxonium_5x5_periodic import PeriodicGraph
from supergrad.scgraph.graph_transmon_1d import Tmon1D

__all__ = ['SCGraph', 'MPCFluxonium1D', 'PeriodicGraph', 'Tmon1D']



import haiku as hk
import numpy as np
import jax.numpy as jnp

from supergrad.helper.helper import Helper
from supergrad.time_evolution import sesolve, sesolve_final_states_w_basis_trans
from supergrad.utils.utility import create_state_init, tensor
from supergrad.time_evolution.single_qubit_compensation import SingleQubitCompensation
from supergrad.scgraph.graph import SCGraph


class Evolve(Helper):
    """Helper for constructing time-evolution computing functions based on the
    graph which contains all information about qubits and pulses. The functions
    constructed this way are pure and can be transformed by Jax.

    Args:
        graph (SCGraph): The graph containing both Hamiltonian and control
            parameters.
        truncated_dim (int):  Desired dimension of the truncated subsystem.
            Note that this applies to all qubits on the graph. One could
            set local configuration for each qubit in the `graph.nodes['qubit']['arguments']`
            to allow different dimension.
        add_random (bool): If true, will add random deviations to the device
            parameters according to the graph.
        share_params (bool): Share device parameters between the qubits that
            have the same shared_param_mark. This is used only for gradient
            computation. One must define `shared_param_mark` in the
            `graph.nodes['qubit']['shared_param_mark']`.
        unify_coupling (bool): Let all couplings in the quantum system be the same.
            TODO: if set to true, which coupling will be used to do the computation?
        coupler_subsystem: Qubits which we set to `|` 0> initially and at the end.
            TODO: make this more general.
        compensation_option: Set single qubit compensation strategy, should be in
            ['no_comp', 'only_vz', 'arbit_single']. 'no_comp' means we do no
            compenstaion. 'only_vz' means we will do single-qubit Z-rotation
            before and after the time evolution. 'arbit_single' means we will do
            arbitrary single-qubit rotation before and after the time evolution.
        solver: the type of time evolution solver, should be in ['ode_expm', 'odeint'].
        options: the arguments will be passed to solver.
            See `supergrad.time_evolution.sesolve`.
    """

    def __init__(self,
                 graph,
                 truncated_dim=5,
                 add_random=True,
                 share_params=False,
                 unify_coupling=False,
                 compensation_option='no_comp',
                 coupler_subsystem=[],
                 solver='ode_expm',
                 options={
                     'astep': 2000,
                     'trotter_order': 1
                 },
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.graph: SCGraph = graph
        self.truncated_dim = truncated_dim
        self.add_random = add_random
        self.share_params = share_params
        self.unify_coupling = unify_coupling
        self.coupler_subsystem = coupler_subsystem
        if compensation_option not in ['no_comp', 'only_vz', 'arbit_single']:
            raise NotImplementedError(
                f'Strategy {compensation_option} has not been implemented.')
        self.compensation_option = compensation_option
        if solver not in ['ode_expm', 'odeint']:
            raise NotImplementedError(
                f'Solver {solver} has not been implemented.')
        self.solver = solver
        self.options = options
        self.psi_list = None

    def _init_quantum_system(self):
        self.hilbertspace = self.graph.convert_graph_to_quantum_system(
            add_random=self.add_random,
            share_params=self.share_params,
            unify_coupling=self.unify_coupling,
            truncated_dim=self.truncated_dim,
            **self.kwargs)
        self.hamiltonian_component_and_pulseshape, self.pulse_endtime = self.graph.convert_graph_to_pulse_lst(
            self.hilbertspace, modulate_wave=True, **self.kwargs)
        if self.compensation_option in ['only_vz', 'arbit_single']:
            sqc = SingleQubitCompensation(self.graph, self.compensation_option,
                                          self.coupler_subsystem)
            self.pre_unitaries, self.post_unitaries = sqc.create_unitaries()

    @property
    def pulse_params(self):
        """The pulse parameters dictionary."""
        pulse_params = self.graph.convert_graph_to_pulse_parameters_haiku()
        if self.compensation_option != 'no_comp':
            # Add initial compensation if not provided in the graph
            initial_comp = self.graph.convert_graph_to_comp_initial_guess(
                self.compensation_option)
            pulse_params = hk.data_structures.merge(initial_comp, pulse_params)
        return pulse_params

    @property
    def all_params(self):
        """The static parameters dictionary."""
        all_params = self.graph.convert_graph_to_parameters_haiku(
            self.share_params, self.unify_coupling)
        if self.compensation_option != 'no_comp':
            # Add initial compensation if not provided in the graph
            initial_comp = self.graph.convert_graph_to_comp_initial_guess(
                self.compensation_option)
            all_params = hk.data_structures.merge(initial_comp, all_params)
        return all_params

    def _prepare_initial_states(self, psi_list=None):
        """Prepare the initial states for time evolution.

        Args:
            psi_list (list, optional): The list of states for evolution. If None,
                all the states in computation basis will be used.
        """
        if psi_list is None:
            states_config = []
            for i, node in enumerate(self.graph.sorted_nodes):
                if node in self.coupler_subsystem:
                    states_config.append([i, 1])
                else:
                    states_config.append([i, 2])
            self.psi_list, _ = create_state_init(self.dims, states_config)
        else:
            self.psi_list = psi_list

    def construct_hamiltonian_and_pulseshape(self):
        """Constructing the Hamiltonian and pulseshape for time evolution.

        Returns:
            (KronObj, list, float): The static Hamiltonian, the list of pair
                containing the time-dependent components of the Hamiltonian and
                corresponding pulse shape, and the maximum length of the pulse.
        """
        return self.hilbertspace.idling_hamiltonian_in_prod_basis(
        ), self.hamiltonian_component_and_pulseshape, self.pulse_endtime

    def _construct_compensation_function(self):
        """Constructing the compensation matrix for virtual compensation.

        Returns:
            Callable: the function that add compensation before and after the
                time evolution unitary.
        """
        if self.compensation_option in ['only_vz', 'arbit_single']:
            pre_u = tensor(*self.pre_unitaries)
            post_u = tensor(*self.post_unitaries)
            if self.compensation_option == 'arbit_single':
                return lambda sim_u: post_u @ sim_u @ pre_u
            else:
                return lambda sim_u: (post_u * sim_u.T).T * pre_u
        else:
            return None

    def eigen_basis(self,
                    transform_matrix=None,
                    psi_list=None,
                    _remove_compensation=False,
                    **kwargs):
        """Running the time evolution in the eigenbasis.

        Args:
            transform_matrix: pre-computed transform matrix,
                using when design parameters is not optimized.
            states: the list of states for evolution. If None, all the states
                in computation basis will be used.
            _remove_compensation: whether to remove the compensation(for debug use).
            kwargs: keyword arguments will be passed to `supergrad.time_evolution.sesolve`
        """
        if transform_matrix is None:
            u_to_eigen = self.hilbertspace.compute_transform_matrix()
        else:
            u_to_eigen = transform_matrix

        # Single qubit product basis
        # construct static hamiltonian
        ham_static = self.hilbertspace.idling_hamiltonian_in_prod_basis()

        ham = [ham_static, *self.hamiltonian_component_and_pulseshape]
        self._prepare_initial_states(psi_list)
        self.tlist = [0, self.pulse_endtime]
        sim_u = sesolve_final_states_w_basis_trans(ham,
                                                   self.psi_list,
                                                   self.tlist,
                                                   transform_matrix=u_to_eigen,
                                                   solver=self.solver,
                                                   options=self.options,
                                                   **kwargs)
        if not _remove_compensation:
            comp_func = self._construct_compensation_function()
            if comp_func is not None:
                sim_u = comp_func(sim_u)
        return sim_u

    def product_basis(self,
                      psi_list=None,
                      _remove_compensation=False,
                      **kwargs):
        """Running the time evolution in the product basis.

        Args:
            psi_list: the list of states for evolution. If None, all the states
                in computation basis will be used.
            _remove_compensation: whether to remove the compensation(for debug use).
            kwargs: keyword arguments will be passed to `supergrad.time_evolution.sesolve`
        """
        # Single qubit product basis
        # construct static hamiltonian
        ham_static = self.hilbertspace.idling_hamiltonian_in_prod_basis()

        ham = [ham_static, *self.hamiltonian_component_and_pulseshape]
        self._prepare_initial_states(psi_list)
        self.tlist = [0, self.pulse_endtime]
        sim_u = sesolve_final_states_w_basis_trans(ham,
                                                   self.psi_list,
                                                   self.tlist,
                                                   transform_matrix=None,
                                                   solver=self.solver,
                                                   options=self.options,
                                                   **kwargs)
        if not _remove_compensation:
            comp_func = self._construct_compensation_function()
            if comp_func is not None:
                sim_u = comp_func(sim_u)
        return sim_u

    def eigen_basis_trajectory(self,
                               tlist=None,
                               psi_list=None,
                               transform_matrix=None,
                               **kwargs):
        """Computing the time evolution trajectory in the eigen basis.

        Args:
            tlist: list of time steps.
            psi_list: the list of states for evolution. If None, all the states
                in computation basis will be used.
            transform_matrix: pre-computed transform matrix,
                using when design parameters is not optimized.
            kwargs: keyword arguments will be passed to `supergrad.time_evolution.sesolve`
        """
        if transform_matrix is None:
            u_to_eigen = self.hilbertspace.compute_transform_matrix()
        else:
            u_to_eigen = transform_matrix

        # Single qubit product basis
        # construct static hamiltonian
        ham_static = self.hilbertspace.idling_hamiltonian_in_prod_basis()

        ham = [ham_static, *self.hamiltonian_component_and_pulseshape]

        transform_matrix = u_to_eigen
        self._prepare_initial_states(psi_list)
        psi_list = transform_matrix @ self.psi_list

        if tlist is None:
            self.tlist = [0, self.pulse_endtime]
        else:
            self.tlist = tlist
        # Change astep as it is used between each steps
        options = self.options.copy()
        options["astep"] = options["astep"] // (len(self.tlist) - 1)
        res = sesolve(ham,
                      psi_list,
                      self.tlist,
                      solver=self.solver,
                      options=options)
        # Replace the last dimension with initial states
        # Now [time, comp, psi_init]
        res = jnp.swapaxes(res, 0, 3)[0]
        states = jnp.conj(transform_matrix).T @ res

        # Extract the states in the computational space
        pop = jnp.abs(psi_list).real**2
        tuple_ar_ix = tuple(np.argmax(pop, axis=1).flatten())

        return states, states[:, tuple_ar_ix, :]

    def product_basis_trajectory(self, tlist=None, psi_list=None, **kwargs):
        """Computing the time evolution trajectory in the product basis.

        Args:
            tlist: list of time steps.
            kwargs: keyword arguments will be passed to `supergrad.time_evolution.sesolve`
        """
        # Single qubit product basis
        # construct static hamiltonian
        ham_static = self.hilbertspace.idling_hamiltonian_in_prod_basis()

        ham = [ham_static, *self.hamiltonian_component_and_pulseshape]
        self._prepare_initial_states(psi_list)
        psi_list = self.psi_list

        if tlist is None:
            self.tlist = [0, self.pulse_endtime]
        else:
            self.tlist = tlist
        # Change astep as it is used between each steps
        options = self.options.copy()
        options["astep"] = options["astep"] // (len(self.tlist) - 1)
        res = sesolve(ham,
                      psi_list,
                      self.tlist,
                      solver=self.solver,
                      options=options)
        # Replace the last dimension with initial states
        # Now [time, comp, psi_init]
        states = jnp.swapaxes(res, 0, 3)[0]

        # Extract the states in the computational space
        pop = jnp.abs(psi_list).real**2
        tuple_ar_ix = tuple(np.argmax(pop, axis=1).flatten())

        return states, states[:, tuple_ar_ix, :]



import numpy as np
import jax
from qiskit_dynamics import Signal
import haiku as hk


def to_qiskit_static_hamiltonian(superObj):
    """Convert the static Hamiltonian to qiskit format.

    Args:
        superObj: The static Hamiltonian in `KronObj` or `LindbladObj` format.
    Returns:
        The static Hamiltonian array in qiskit format.
    """
    return superObj.full()


def to_qiskit_drive_hamiltonian(hamiltonian_component_and_pulseshape):
    """Convert the drive Hamiltonian to qiskit format.

    Args:
        hamiltonian_component_and_pulseshape: The list of pair containing the
            time-dependent components of the Hamiltonian and corresponding pulse
            shape.
    Returns:
        (list, list)
        The list of Hamiltonian operators and the list of signal in qiskit format.
    """
    drive_ham_list = []
    signals_list = []
    for drive, create_pulse in hamiltonian_component_and_pulseshape:
        # convert pulse as qiskit signal
        pulse = create_pulse.__self__
        drive_ham_list.append(drive.full())
        carrier_freq = pulse.omega_d / 2 / np.pi

        def signal_closure(pulse):
            # Instantiation same envelope function

            @hk.without_apply_rng
            @hk.transform
            def envelope(t):
                new_pulse = type(pulse)(modulate_wave=False, name=pulse.name)
                return new_pulse.create_envelope_pulse(t)

            # get parameters template
            rng = jax.random.PRNGKey(0)
            params = envelope.init(rng, 0.)
            # filling params by existed pulse
            existed_params = vars(pulse)
            existed_params = hk.data_structures.filter(
                lambda module_name, name, value: name in params[pulse.name].
                keys(), {pulse.name: existed_params})
            params = hk.data_structures.merge(params, existed_params)
            return lambda t: envelope.apply(params, t)

        signals_list.append(
            Signal(signal_closure(pulse), carrier_freq, pulse.phase,
                   pulse.name))
    return drive_ham_list, signals_list



from supergrad.version import __version__ as __version__

from jax import config

config.update("jax_enable_x64", True)
# enable fp64 for better numerical stability

from supergrad.utils import identity_wrap as identity_wrap
from supergrad.utils import tensor as tensor
from supergrad.utils import basis as basis
from supergrad.utils import permute as permute

from supergrad.quantum_system import KronObj as KronObj
from supergrad.quantum_system import LindbladObj as LindbladObj

from supergrad.helper import Helper as Helper
from supergrad.helper import Spectrum as Spectrum
from supergrad.helper import Evolve as Evolve

from supergrad.time_evolution import sesolve as sesolve
from supergrad.time_evolution import mesolve as mesolve



from functools import partial
import operator

import jax
import jax.numpy as jnp
from jax.scipy.linalg import expm, eigh
from jax.numpy import interp
from jax.tree_util import tree_map
from jax.experimental.ode import odeint

import supergrad
from supergrad.quantum_system import KronObj, LindbladObj
from supergrad.utils.progress_bar import scan_tqdm


def _check_compatibility_mode(compatibility_mode, dim):
    if compatibility_mode:
        return jnp.eye(dim)
    else:
        return 0.


def _expm_mul_y(mat, y=None, *args, **kwargs):
    """Combine matrix exponentiation with matrix-vector multiplication.

    Args:
        mat: the matrix
        y: the vector be multiplied by the matrix on its left
        args: arguments be passed to `KronObj.expm`
        kwargs: the keyword arguments be passed to `KronObj.expm`
    """
    if isinstance(mat, (KronObj, LindbladObj)):
        return mat.expm(y, *args, **kwargs)
    else:
        return expm(mat) @ y


def ode_expm(func,
             y0,
             ts,
             *args,
             astep=100,
             trotter_order=None,
             progress_bar=False,
             custom_vjp=True,
             pb_fwd_ad=False,
             compatibility_mode=False):
    """ODE solver using the matrix exponentiation for the propagators at each time
    step.

    Args:
        func: function to evaluate the time derivative of the solution `y` at
            time `t` as `func(y, t, *args)`, producing the same shape/structure
            as `y0`.
        y0: array or pytree of arrays representing the initial value for the state.
        ts: array of float times for evaluation, like `jnp.linspace(0., 10., 101)`,
            in which the values must be strictly increasing.
        *args: tuple of additional arguments for `func`, which must be arrays
            scalars, or (nested) standard Python containers (tuples, lists,
            dicts, namedtuples, i.e. pytrees) of those types.
        astep: int, absolute number of steps to take for each timepoint (optional).
        trotter_order (complex int): the order of suzuki-trotter decomposition.
            The following arguments are supported,
            a) `None`, calculating matrix exponentiation without trotter decomposition
            b) `1`, first order trotter decomposition
            c) `2`, second order trotter decomposition
            d) `4`, 4th order real decomposition
            e) `4j`, 4th order complex decomposition
        progress_bar (bool) : whether to display a progress bar (optional).
        custom_vjp (string) : choose custom automatic differentiation `VJP` rule.
            Default is `None`, which means using the JAX framework to derive `VJP`.
            The following arguments are supported:
            a) True or `LCAM` (Recommended) : using the local continuous adjoint
            method to compute the inverse time evolution for both state and
            adjoint state.
            b) `CAM` : using the continuous adjoint method.
        pb_fwd_ad (bool) : whether to config progress bar as forward mode
            automatic differentiation.
        compatibility_mode (bool) : whether to use the compatible mode for `func`.
            We disable compatible mode to reduce computational cost when the evolution
            operator is not depend on `y`, so we could let `y` equals to 0 in the `func`.


    Returns:
        Values of the solution `y` (i.e. integrated system values) at each time
        point in `t`, represented as an array (or pytree of arrays) with the
        same shape/s

    """
    # convert the func to closure to make it custom_vjp compatible
    converted, consts = jax.custom_derivatives.closure_convert(
        func, y0, ts[0], *args)
    y_eye = _check_compatibility_mode(compatibility_mode, y0.shape[0])
    # compute optimal contraction path
    evo_term = func(y_eye, ts[0], *args)
    if trotter_order is not None and isinstance(evo_term,
                                                (KronObj, LindbladObj)):
        tn_expr = evo_term.compute_contraction_path(trotter_order=trotter_order)
    else:
        tn_expr = None
    if custom_vjp is not None:
        assert not pb_fwd_ad, 'Forward mode automatic differentiation is not supported.'
        _custom_ode_expm = jax.custom_vjp(_ode_expm,
                                          nondiff_argnums=(0, 1, 2, 3, 4, 5, 6))
        if custom_vjp is True or custom_vjp == 'LCAM':
            _custom_ode_expm.defvjp(_ode_expm_fwd, _ode_expm_rev)
        elif custom_vjp == 'CAM':
            _custom_ode_expm.defvjp(_ode_expm_fwd, _ode_expm_rev_continuous)
        else:
            raise ValueError('Invalid custom_vjp type.')
        return _custom_ode_expm(converted, astep, trotter_order, tn_expr,
                                progress_bar, False, compatibility_mode, y0, ts,
                                *args, *consts)
    else:
        return _ode_expm(converted, astep, trotter_order, tn_expr, progress_bar,
                         pb_fwd_ad, compatibility_mode, y0, ts, *args, *consts)


@partial(jax.jit, static_argnums=(0, 1, 2, 3, 4, 5, 6))
def _ode_expm(func, astep, trotter_order, tn_expr, progress_bar, pb_fwd_ad,
              compatibility_mode, y0, ts, *args):
    """The private function of the matrix exponentiation ODE solver.
    Operations that only depend on static arguments will be constant-folded in
    Python (during tracing), and so the corresponding argument values can be
    any Python object.
    """
    if isinstance(y0, KronObj):
        # convert y0 to the dense form
        y0 = y0.full()

    def scan_func(carry, target_t):
        """Evolve the propagator K to evaluation state at time point in ts."""

        def body_fun(state, next_nest_t):
            """Evolution of the discrete time sequence."""
            nest_y, nest_t = state
            # it's reasonable to choose the middle point of every time slice
            evo_term = func(y_eye, (nest_t + next_nest_t) / 2, *args)
            # evolution a timestep for each timepoint
            next_nest_y = _expm_mul_y(evo_term * (next_nest_t - nest_t), nest_y,
                                      trotter_order, tn_expr)
            # normalize exp(-iHt)
            # if nest_y.shape[1] == 1:
            #     norm = jnp.linalg.norm(next_nest_y)
            #     next_nest_y = next_nest_y / norm**2
            # elif nest_y.shape[0] == nest_y.shape[1]:
            #     norm = jnp.trace(next_nest_y)
            #     next_nest_y = next_nest_y / norm
            # else:
            #     raise ValueError('Shape is not compatible.')
            carry = [next_nest_y, next_nest_t]
            return carry, next_nest_t

        y, t = carry
        # decompose the total propagator K to the propagator U for the short time
        nest_ts = jnp.linspace(t, target_t, astep)
        state = [y, t]
        if progress_bar:
            wrapped_body_fun, wrapped_nest_ts = scan_tqdm(body_fun,
                                                          nest_ts[1:],
                                                          fwd_ad=pb_fwd_ad)
            final_state, _ = jax.lax.scan(wrapped_body_fun, state,
                                          wrapped_nest_ts)
        else:
            final_state, _ = jax.lax.scan(body_fun, state, nest_ts[1:])
        next_y = final_state[0]

        return final_state, next_y

    # initialize evolution
    init_carry = [y0, ts[0]]
    # check compatible mode
    y_eye = _check_compatibility_mode(compatibility_mode, y0.shape[0])
    _, ys = jax.lax.scan(scan_func, init_carry, ts[1:])

    return jnp.concatenate((y0[None], ys))


def _ode_expm_fwd(func, astep, trotter_order, tn_expr, progress_bar, pb_fwd_ad,
                  compatibility_mode, y0, ts, *args):
    """Forward pass of the matrix exponentiation ODE solver."""
    yf = _ode_expm(func, astep, trotter_order, tn_expr, progress_bar, pb_fwd_ad,
                   compatibility_mode, y0, ts, *args)
    return yf, (yf, ts, args)


def _ode_expm_rev(func, astep, trotter_order, tn_expr, progress_bar, pb_fwd_ad,
                  compatibility_mode, res, ys_bar):
    """Reverse pass of the matrix exponentiation ODE solver.
    Inverse the origin time evolution augmented with vjp_y, This approach only
    works for the Hermitian Hamiltonian .
    """

    def evolution_step(y, t, next_t, *args):
        evo_term = func(y_eye, (next_t + t) / 2, *args).dag()
        next_y = _expm_mul_y(evo_term * (t - next_t), y, trotter_order, tn_expr)
        return next_y

    def aug_evolution_step(y_bar, t, next_t, *args):
        evo_term = func(y_eye, (next_t + t) / 2, *args).conjugate()
        next_y = _expm_mul_y(evo_term * (next_t - t), y_bar, trotter_order,
                             tn_expr)
        return next_y

    def aug_dynamics(augmented_state, t, *args):
        """Backward ODE about vjp_t and vjp_args."""
        y, y_bar, *_ = augmented_state
        y_dot, vjpfun = jax.vjp(
            lambda _y, _t, _args: func(y_eye, _t, *_args) @ _y, y, t, args)
        return (-y_dot, *vjpfun(y_bar))

    def scan_func(carry, idx):
        """Evolve the propagator K to evaluation state at time point in ts."""

        def body_fun(state, next_nest_t):
            """Evolution of the discrete time sequence."""
            nest_y, nest_y_bar, nest_t_bar, nest_args_bar, nest_t = state
            # compute the vjp of hamiltonian function, back-propagation vjp_y
            next_nest_y, vjpfun = jax.vjp(evolution_step, nest_y, nest_t,
                                          next_nest_t, *args)
            # inverse the time evolution
            next_nest_y_bar = aug_evolution_step(nest_y_bar, nest_t,
                                                 next_nest_t, *args)

            # a) chain rule (accuracy but high cost)
            _, rev_t_bar, next_nest_t_bar, *next_nest_args_bar = vjpfun(
                next_nest_y_bar)
            next_nest_t_bar = nest_t_bar - rev_t_bar - next_nest_t_bar
            next_nest_args_bar = jax.tree_util.tree_map(lambda x, y: x - y,
                                                        nest_args_bar,
                                                        next_nest_args_bar)

            # b) compute augmented dynamics (trapezoid rule, low accuracy)
            # aug_state_l = [nest_y, nest_y_bar, nest_t_bar, nest_args_bar]
            # aug_state_h = [
            #     next_nest_y, next_nest_y_bar, nest_t_bar, nest_args_bar
            # ]
            # _, _, t_dy_l, args_dy_l = aug_dynamics(aug_state_l,
            #                                        (nest_t + next_nest_t) / 2,
            #                                        *args)
            # _, _, t_dy_h, args_dy_h = aug_dynamics(aug_state_h,
            #                                        (nest_t + next_nest_t) / 2,
            #                                        *args)
            # # cumulative vjp_t and vjp_args
            # next_nest_t_bar = nest_t_bar - (t_dy_l + t_dy_h) / 2 * (nest_t -
            #                                                         next_nest_t)
            # next_nest_args_bar = jax.tree_util.tree_map(
            #     lambda x, y, z: x + (y + z) / 2 * (nest_t - next_nest_t),
            #     nest_args_bar, list(args_dy_l), list(args_dy_h))
            # rev_t_bar = next_nest_t_bar

            carry = [
                next_nest_y, next_nest_y_bar, next_nest_t_bar,
                next_nest_args_bar, next_nest_t
            ]
            return carry, -rev_t_bar

        _, y_bar, t0_bar, args_bar = carry
        nest_ts = jnp.linspace(ts[idx], ts[idx - 1], astep)
        # Compute effect of moving measurement time
        # `t_bar` should not be complex as it represents time
        mea_t_bar = jnp.dot((func(y_eye, ts[idx], *args) @ ys[idx]).T,
                            ys_bar[idx]).real.reshape(-1)[0]
        t0_bar = t0_bar - mea_t_bar
        # state is [y, y_bar, t_bar, args_bar, t]
        state = [ys[idx], y_bar, t0_bar, args_bar, ts[idx]]
        if progress_bar:
            wrapped_body_fun, wrapped_nest_ts = scan_tqdm(body_fun,
                                                          nest_ts[1:],
                                                          fwd_ad=True)
            final_state, _ = jax.lax.scan(wrapped_body_fun, state,
                                          wrapped_nest_ts)
        else:
            final_state, _ = jax.lax.scan(body_fun, state, nest_ts[1:])
        # Add gradient from current output
        final_state[1] += ys_bar[idx - 1]
        # check symmetry by compare cached y with inverse computed one
        # jax.debug.print(
        #     f'Check symmetry: {jnp.allclose(final_state[0], ys[idx - 1])}')
        # update carry
        return final_state[:-1], mea_t_bar

    ys, ts, args = res
    # check compatible mode
    y_eye = _check_compatibility_mode(compatibility_mode, ys[-1].shape[0])
    # carry is [y, y_bar, t_bar, args_bar]
    # compute initial states
    init_carry = [
        ys[-1], ys_bar[-1], 0.,
        jax.tree_util.tree_map(jnp.zeros_like, list(args))
    ]
    # inverse the time evolution for adjoint state
    (_, y_bar, t0_bar,
     args_bar), rev_ts_bar = jax.lax.scan(scan_func, init_carry,
                                          jnp.arange(len(ts) - 1, 0, -1))
    ts_bar = jnp.concatenate([jnp.array([t0_bar]), rev_ts_bar[::-1]])
    return y_bar, ts_bar, *args_bar


def _ode_expm_rev_continuous(func, astep, trotter_order, tn_expr, progress_bar,
                             pb_fwd_ad, compatibility_mode, res, ys_bar):
    """Reverse pass of the matrix exponentiation ODE solver using continuous
    adjoint method.
    """

    def aug_dynamics(augmented_state, t, *args):
        """Original system augmented with vjp_y, vjp_t and vjp_args."""
        y, y_bar, *_ = augmented_state
        # `t` here is negatice time, so we need to negate again to get back to
        # normal time. See the `odeint` invocation in `scan_fun` below.
        y_dot, vjpfun = jax.vjp(
            lambda _y, _t, _args: func(y_eye, _t, *_args) @ _y, y, -t, args)
        return (-y_dot, *vjpfun(y_bar))

    ys, ts, args = res
    # check compatible mode
    y_eye = _check_compatibility_mode(compatibility_mode, ys[-1].shape[0])
    y_bar = ys_bar[-1]
    ts_bar = []
    t0_bar = 0.

    def scan_fun(carry, i):
        y_bar, t0_bar, args_bar = carry
        # Compute effect of moving measurement time
        # `t_bar` should not be complex as it represents time
        t_bar = jnp.dot((func(y_eye, ts[i], *args) @ ys[i]).T,
                        ys_bar[i]).real.reshape(-1)[0]
        t0_bar = t0_bar - t_bar
        # Run augmented system backwards to previous observation

        _, y_bar, t0_bar, args_bar = odeint(aug_dynamics,
                                            (ys[i], y_bar, t0_bar, args_bar),
                                            jnp.array([-ts[i], -ts[i - 1]]),
                                            *args,
                                            rtol=1e-10,
                                            atol=1e-10)
        # the tolerance of the ODE solve is set to 1e-10 to ensure the accuracy
        # it's hard to determine tolerance based on `astep` and `trotter_order`
        y_bar, t0_bar, args_bar = tree_map(operator.itemgetter(1),
                                           (y_bar, t0_bar, args_bar))
        # Add gradient from current output
        y_bar = y_bar + ys_bar[i - 1]
        return (y_bar, t0_bar, args_bar), t_bar

    init_carry = (ys_bar[-1], 0., tree_map(jnp.zeros_like, args))
    (y_bar, t0_bar,
     args_bar), rev_ts_bar = jax.lax.scan(scan_fun, init_carry,
                                          jnp.arange(len(ts) - 1, 0, -1))
    ts_bar = jnp.concatenate([jnp.array([t0_bar]), rev_ts_bar[::-1]])
    return (y_bar, ts_bar, *args_bar)


class EvoElement:
    """
    Internal type used to represent the time-dependent parts of a class.

    Availables "types" are

    1. ``jnp.ndarray``
    2. function

    Args:
        oper: the time-independent hamiltonian
        get_coeff: a callable that take (t, args) and return the coeff at that t
        coeff: The coeff as a ndarray as provided by the user.
        _type: type of `EvoElement`
    """

    def __init__(self, oper, get_coeff, coeff: jnp.ndarray, _type: str):
        self.oper = oper
        self.get_coeff = get_coeff
        self.coeff = coeff
        self._type = _type

    @classmethod
    def make(cls, list_):
        return cls(*list_)

    def __getitem__(self, i):
        if i == 0:
            return self.oper
        if i == 1:
            return self.get_coeff
        if i == 2:
            return self.coeff
        if i == 3:
            return self._type


class _InterpolateWrapper:
    """Using JAX interpolate since ODE solver only accept
    linearly distributed tlist.
    """

    def __init__(self, tlist, coeff, args=None):
        self.coeff = coeff
        self.tlist = tlist

    def __call__(self, t, args={}):
        return interp(t, self.tlist, self.coeff, left=0., right=0.)


def _parse_hamiltonian(q_object, tlist: jnp.ndarray, args: dict,
                       diag_ops: bool):
    """Prepare the time dependent Hamiltonian for the solver.

    Args:
        q_object (list, array):
            The time-dependent description of the quantum object.  This is of
            the same format as the first parameter to the general ODE solvers;
            in general, it is a list of ``[ndarray, time_dependence]`` pairs
            that are summed to make the whole object.  The ``time_dependence``
            can be any of the formats discussed in the previous section.

        args (dict, optional):
            Mapping of ``{str: object}``, discussed in greater detail above.
            The strings can be any valid Python identifier, and the objects are
            of the consumable types.  See the previous section for details on
            the "magic" names used to access solver internals.

        tlist (list, array):
            List of the times any numpy-array coefficients describe.  This is
            used only in at least one of the time dependence in ``H`` is given
            in Numpy-array format.  The times must be sorted, but need not be
            equidistant.  Values inbetween will be interpolated.

    Returns:
        func(Callable): Return a single `ndarray` at the given time `t`.
    """

    def _td_op_type(element):
        if isinstance(element, (KronObj, jnp.ndarray)):
            return 0
        try:
            op, td = element
        except (TypeError, ValueError) as exc:
            raise TypeError('Incorrect q_object specification.') from exc
        if callable(td):
            out = 1
        elif isinstance(td, (KronObj, jnp.ndarray)):
            if td.shape != tlist.shape:
                raise ValueError('Time sequences are not compatible.')
            out = 3
        else:
            raise TypeError('Incorrect q_object specification.')
        return out

    def _td_format_check(q_object):
        if isinstance(q_object, (KronObj, jnp.ndarray)):
            return 0
        if isinstance(q_object, list):
            return [_td_op_type(element) for element in q_object]
        raise TypeError('Incorrect q_object specification.')

    # Initialize parameters
    const = False
    cte = None
    tlist = jnp.asarray(tlist)

    # Attempt to determine if a 2-element list is a single, time-dependent
    # operator, or a list with 2 possibly time-dependent elements.
    if isinstance(q_object, list) and len(q_object) == 2:
        try:
            # Test if parsing succeeds on this as a single element.
            _td_op_type(q_object)
            q_object = [q_object]
        except (TypeError, ValueError):
            pass

    op_type = _td_format_check(q_object)
    ops = []

    if isinstance(op_type, int):
        if op_type == 0:
            cte = q_object
            const = True
    else:
        for type_, op in zip(op_type, q_object):
            if type_ == 0:
                if cte is None:
                    cte = op
                else:
                    cte += op
            elif type_ == 1:
                ops.append(EvoElement(op[0], op[1], op[1], 'func'))
            elif type_ == 3:
                ops.append(
                    EvoElement(op[0],
                               _InterpolateWrapper(tlist, op[1], args=args),
                               jnp.copy(op[1]), tlist))

        if not ops:
            const = True

    if diag_ops and isinstance(cte, KronObj):
        cte = cte.diagonalize_operator()
    eig_info = []
    if diag_ops:
        # diagonalize time dependent hamiltonian
        for part in ops:
            part_eig_info = []
            if isinstance(part.oper, KronObj):
                for nest_mat in part.oper.data:
                    mat = supergrad.tensor(*nest_mat)
                    part_eig_info.append(eigh(mat))
            eig_info.append(part_eig_info)

    # Define a callable Hamiltonian function.
    def hamiltonian(t, imul=1, args={}):
        """Return a single `ndarray` at the given time ``t``.

        Args:
            t: the given time.
            imul: the term should multiply to hamiltonian.
        """
        op_t = imul * cte
        if not const:
            if diag_ops:
                for partid, part in enumerate(ops):
                    # diagonalize time dependent hamiltonian
                    part_eig_info = eig_info[partid]
                    if isinstance(part.oper, KronObj):
                        # iteration through ham_td
                        for idx, sub_kronobj in enumerate(part.oper):
                            eig_val, eig_vec = part_eig_info[idx]
                            # construct KronObj in diag unitary method
                            op_t += KronObj(
                                [eig_val],
                                sub_kronobj.dims,
                                sub_kronobj.locs[0],
                                diag_unitary=eig_vec,
                            ) * part.get_coeff(t, args) * imul
                    else:
                        mat = part.oper
                        eig_val, eig_vec = eigh(mat)
                        # construct KronObj in diag unitary method
                        op_t += KronObj(
                            [eig_val],
                            diag_unitary=eig_vec,
                        ) * part.get_coeff(t, args) * imul
            else:
                for part in ops:
                    op_t += part.oper * part.get_coeff(t, args) * imul

        return op_t

    return hamiltonian



import copy
from typing import List
import warnings
import jax
import jax.numpy as jnp
import networkx as nx
from networkx.classes.reportviews import EdgeView, NodeView

from supergrad.time_evolution.pulseshape import (PulseCosine, PulseTrapezoid,
                                                 PulseCosineRamping, PulseTanh,
                                                 PulseErf, PulseGaussian)
from supergrad.utils.utility import identity_wrap, convert_to_haiku_dict

from supergrad.quantum_system.artificial_atom import (Fluxonium, Transmon,
                                                      Resonator)
from supergrad.quantum_system.interaction import (InteractingSystem,
                                                  InteractionTerm,
                                                  parse_interaction)

# please register module in the following dictionary for graph converting
artificial_atom_dict = {
    'fluxonium': Fluxonium,
    'transmon': Transmon,
    'resonator': Resonator
}
pulse_shape_dict = {
    'trapezoid': PulseTrapezoid,
    'cos': PulseCosine,
    'rampcos': PulseCosineRamping,
    'tanh': PulseTanh,
    'erf': PulseErf,
    'gaussian': PulseGaussian
}


def parse_pre_comp_name(q):
    """Parse the previous compensation name. The compensation name consist of
    three parts separated by underscores '_', the third part is nodes' name.

    Args:
        q (str): the node name
    """
    key = '_'.join(['pre', 'comp', q])
    return key


def parse_post_comp_name(q):
    """Parse the post compensation name. The compensation name consist of
    three parts separated by underscores '_', the third part is nodes' name.

    Args:
        q (str): the node name
    """
    key = '_'.join(['post', 'comp', q])
    return key


def _parse_edges_name(k, q1, q2):
    """Parse interaction name. The compensation name consist of four parts
    separated by underscores '_'. The first and second part could be either
    'capacitive_coupling' or 'inductive_coupling', the third and fourth
    part are sorted qubits' name.

    Args:
        k (str): coupling type in ['capacitive_coupling', 'inductive_coupling']
        q1 (str): a node name
        q2 (str): a node name
    """
    key = '_'.join(sorted([q1, q2]))  # use string intrinsic order
    key = '_'.join([k, key])
    return key


def _parse_pulse_name(q: str, name: str, pulse: str):
    """Parse pulseshape name. The compensation name consist of three parts
    separated by underscores '_', they are node name, 'pulse' and the registered
    pulse type, respectively.

    Args:
        q (str): a node name
        name (str): the pulse name
        pulse (str): the registered pulse type
    """
    if name.find("_") != -1:
        raise ValueError("Pulse name should not contain '_'.")
    key = '_'.join([q, name, pulse])
    return key


class SCGraph(nx.Graph):
    """Graph for storing parameters of qubits, their coupling strength
    and the control pulses.
    It is inherited from the `networkx.Graph`, which provides an implementation
    of graph structure.
    """

    def __init__(self, incoming_graph_data=None, **attr):
        super().__init__(incoming_graph_data, **attr)

        # register the global seed and PRNGKey to support random number.
        self._seed = None
        self.key = None

    @property
    def sorted_nodes(self):
        """A list contains all the nodes in deterministic order."""
        return sorted(list(NodeView(self)))

    @property
    def sorted_edges(self):
        """A list contains all the edges in deterministic order."""
        return sorted(list(EdgeView(self)))

    @property
    def seed(self):
        """The seed for random device parameters variance."""
        return self._seed

    @seed.setter
    def seed(self, new_seed):
        if isinstance(new_seed, int):
            self._seed = new_seed
            # update the PRNGKey
            self.key = jax.random.PRNGKey(self._seed)
        else:
            raise ValueError('Please use integer as random seed.')

    def add_lcj_params_variance_to_graph(self, multi_err=0.01, seed=1):
        """Assign (random) values of variance to superconducting qubits' parameters
        ['ec', 'ej', 'el'] in the graph. Note that here we only generate and store
        the multiplicative factors in the graph. They are not applied yet.

        Args:
            multi_err (float): the strength of the multiplicative error.
                More concretely, a different factor 1+ N * multi_error will be
                multiplied to each parameter, where N is sampled from normal
                distribution.
            seed (int): the random seed to generate parameters variance.
        """
        # generate PRNGKey by setter
        self.seed = seed
        for node_name in self.sorted_nodes:
            self.key, subkey = jax.random.split(self.key)
            node = NodeView(self)[node_name]
            params_variance = 1 + jax.random.normal(subkey,
                                                    shape=(3,)) * multi_err
            variance = {}
            for i, item in enumerate(['ec', 'ej', 'el']):
                variance[item] = params_variance[i]
            node['variance'] = variance

    def _convert_graph_to_pulse_parameters(self, node, node_name, params_haiku):
        """Convert a networkX graph to Haiku's parameters dictionary.
        Only for internal use.
        """
        # update pulse parameters
        for k, v in dict(node).items():
            if str(k).startswith('pulse'):
                pulse_type = v.pop('pulse_type')
                key = _parse_pulse_name(node_name, k, pulse_type)
                # parameters tidy up
                for unused_params in [
                        'delay', 'operator_type', 'crosstalk', 'arguments'
                ]:
                    v.pop(unused_params, None)
                params_haiku.update({key: v})
        return params_haiku

    def _convert_graph_to_compensation(self, node, node_name,
                                       compensation_params):
        """Convert a networkX graph to Haiku's parameters dictionary.
        Only for internal use.
        """
        for k, v in dict(node).items():
            if str(k).startswith('compensation'):
                pre_comp_val = v['pre_comp']
                post_comp_val = v['post_comp']
                compensation_params.update(
                    {parse_pre_comp_name(node_name): pre_comp_val})
                compensation_params.update(
                    {parse_post_comp_name(node_name): post_comp_val})
        return compensation_params

    def convert_graph_to_pulse_parameters_haiku(self, load_compensation=True):
        """Convert a networkX graph to Haiku's parameters dictionary.

        Args:
            load_compensation: whether to load the compensation parameters.
        """
        graph = copy.deepcopy(self)
        params_haiku = {}
        # update artificial atom parameters
        compensation_params = {}
        for node_name in self.sorted_nodes:
            node: dict = graph.nodes[node_name]
            self._convert_graph_to_pulse_parameters(node, node_name,
                                                    params_haiku)
            self._convert_graph_to_compensation(node, node_name,
                                                compensation_params)
        # update compensation parameters
        if compensation_params and load_compensation:
            params_haiku.update({'single_q_compensation': compensation_params})
        return convert_to_haiku_dict(params_haiku)

    def convert_graph_to_parameters_haiku(self,
                                          share_params=False,
                                          unify_coupling=False,
                                          only_device_params=False):
        """Convert a networkX graph to Haiku's parameters dictionary.

        Args:
            share_params (bool): If True, two sets of parameters will be shared
                according to the shared_param_mark attribute in the graph. This means
                that the shared parameters will remain the same during optimization.
                First, qubits with the same shared_param_mark will have their parameters
                shared. Second, the coupling strengths between qubits of
                type i and j are shared for all such pairs.
            unify_coupling (bool): all coupling sharing the same parameters.
                Valid only if `share_params` is True.
            only_device_params (bool): whether to only load device parameters.
        """
        graph = copy.deepcopy(self)
        params_haiku = {}
        # update artificial atom parameters
        type_list = []
        compensation_params = {}
        for node_name in self.sorted_nodes:
            node: dict = graph.nodes[node_name]
            node_type = node.get('shared_param_mark', node_name)
            # update pulse parameters
            if not only_device_params:
                self._convert_graph_to_pulse_parameters(node, node_name,
                                                        params_haiku)
                self._convert_graph_to_compensation(node, node_name,
                                                    compensation_params)
            if node_type not in type_list or not share_params:
                qubit_params = {}
                attr_dict = dict(node)
                for kw in ['ec', 'ej', 'el', 'phiext']:
                    val = attr_dict.pop(kw, None)
                    if val is not None:
                        qubit_params.update({kw: val})
                params_haiku.update({node_name: qubit_params})
            type_list.append(node_type)
        # update compensation parameters
        if compensation_params and not only_device_params:
            params_haiku.update({'single_q_compensation': compensation_params})
        # update interaction parameters
        edge_type_list = []
        for edge in self.sorted_edges:
            q1, q2 = edge
            attr = graph.edges[edge]
            edge_type = tuple(
                sorted([
                    graph.nodes[node].get('shared_param_mark', node)
                    for node in [q1, q2]
                ]))
            for k, v in dict(attr).items():
                if unify_coupling:
                    key = _parse_edges_name(k, 'all', 'unify')
                elif edge_type_list and unify_coupling:
                    continue
                else:
                    if edge_type in edge_type_list and share_params:
                        continue
                    else:
                        key = _parse_edges_name(k, q1, q2)
                params_haiku.update({key: v})
            edge_type_list.append(edge_type)
        return convert_to_haiku_dict(params_haiku)

    def convert_graph_to_comp_initial_guess(self,
                                            compensation_option='only_vz',
                                            name='single_q_compensation'):
        """Convert a networkX graph to virtual compensation initial guess.

        Args:
            compensation_option: single qubit compensation strategy, should be
                in ['only_vz', 'arbit_single]
            name: name of the single qubit compensation module, default name is
                'single_q_compensation'.
        """
        assert compensation_option in ['only_vz', 'arbit_single']
        shape = [] if compensation_option == 'only_vz' else [3]
        comp_dict = {}
        for node in self.sorted_nodes:
            comp_dict.update({parse_pre_comp_name(node): jnp.zeros(shape)})
            comp_dict.update({parse_post_comp_name(node): jnp.zeros(shape)})
        return {name: comp_dict}

    def convert_graph_to_quantum_system(self,
                                        add_random: bool = True,
                                        share_params: bool = True,
                                        unify_coupling: bool = False,
                                        **kwargs):
        """Convert a networkX graph to SuperGrad quantum system.

        Args:
            add_random: whether adding random device parameters deviations or not
            share_params : If True, two sets of parameters will be shared
                according to the shared_param_mark attribute in the graph. This means
                that the shared parameters will remain the same during optimization.
                First, qubits with the same shared_param_mark will have their parameters
                shared. Second, the coupling strengths between qubits of
                type i and j are shared for all such pairs.
            unify_coupling: all coupling sharing the same parameters. Valid
                only if `share_params` is True.
        """
        # check arguments
        if unify_coupling:
            assert share_params is True
        # convert nodes to artificial atoms
        subsystem_list = []
        type_list = []
        for node_name in self.sorted_nodes:
            node = NodeView(self)[node_name]
            node_type = node.get('shared_param_mark', node_name)
            node_kwargs = copy.deepcopy(kwargs)
            # update global kwargs with graph kwargs
            node_kwargs.update(node.get('arguments', {}))
            if node_type in type_list and share_params:
                type_idx = type_list.index(node_type)
                mirrored_subsystem = copy.deepcopy(subsystem_list[type_idx])
                # reconstruct class
                mirrored_subsystem.name = node_name
                if add_random:
                    var = node.get('variance', None)
                    mirrored_subsystem.add_lcj_params_variance(var)
                subsystem_list.append(mirrored_subsystem)
            else:
                var = None
                if add_random:
                    var = node.get('variance', None)
                subsystem_list.append(artificial_atom_dict[node['system_type']](
                    name=node_name, var=var, **node_kwargs))
            type_list.append(node_type)
        subsystem_name_list = self.sorted_nodes
        interaction_list = []
        edge_type_list = []
        for edge in self.sorted_edges:
            q1, q2 = edge
            attr = EdgeView(self)[edge]
            edge_type = tuple(
                sorted([
                    NodeView(self)[node].get('shared_param_mark', node)
                    for node in [q1, q2]
                ]))
            idx_q1 = subsystem_name_list.index(q1)
            idx_q2 = subsystem_name_list.index(q2)
            if unify_coupling:
                predicate = interaction_list and share_params  # TODO what is predicate
            else:
                predicate = (edge_type in edge_type_list and share_params)

            if predicate:
                if unify_coupling:
                    mirrored_interaction_pair: List[
                        InteractionTerm] = copy.deepcopy(interaction_list[0])
                else:
                    edge_type_idx = edge_type_list.index(edge_type)
                    mirrored_interaction_pair: List[
                        InteractionTerm] = copy.deepcopy(
                            interaction_list[edge_type_idx])
                # reconstruct class
                for mirror in mirrored_interaction_pair:
                    old_name = mirror.name
                    k = '_'.join(old_name.split('_')[0:2])
                    mirror.name = _parse_edges_name(k, q1, q2)
                    if k == 'capacitive_coupling':
                        mirror.operator_list = [
                            subsystem_list[idx_q1].n_operator,
                            subsystem_list[idx_q2].n_operator
                        ]
                    elif k == 'inductive_coupling':
                        mirror.operator_list = [
                            subsystem_list[idx_q1].phi_operator,
                            subsystem_list[idx_q2].phi_operator
                        ]
                interaction_list.append(mirrored_interaction_pair)
            else:
                interaction_pair = []
                if unify_coupling:
                    for k in dict(attr).keys():
                        if k == 'capacitive_coupling':
                            unify_cap_coupling = parse_interaction(
                                op1=subsystem_list[idx_q1].n_operator,
                                op2=subsystem_list[idx_q2].n_operator,
                                name=_parse_edges_name(k, 'all', 'unify'))
                            unify_cap_coupling.name = _parse_edges_name(
                                k, q1, q2)
                            interaction_pair.append(unify_cap_coupling)
                        elif k == 'inductive_coupling':
                            unify_ind_coupling = parse_interaction(
                                op1=subsystem_list[idx_q1].phi_operator,
                                op2=subsystem_list[idx_q2].phi_operator,
                                name=_parse_edges_name(k, 'all', 'unify'))
                            unify_ind_coupling.name = _parse_edges_name(
                                k, q1, q2)
                            interaction_pair.append(unify_ind_coupling)
                else:
                    for k in dict(attr).keys():
                        if k == 'capacitive_coupling':
                            interaction_pair.append(
                                parse_interaction(
                                    op1=subsystem_list[idx_q1].n_operator,
                                    op2=subsystem_list[idx_q2].n_operator,
                                    name=_parse_edges_name(k, q1, q2)))
                        elif k == 'inductive_coupling':
                            interaction_pair.append(
                                parse_interaction(
                                    op1=subsystem_list[idx_q1].phi_operator,
                                    op2=subsystem_list[idx_q2].phi_operator,
                                    name=_parse_edges_name(k, q1, q2)))
                interaction_list.append(interaction_pair)

            edge_type_list.append(edge_type)
        # unflatten interaction list
        inter_list = []
        for inter_pair in interaction_list:
            inter_list.extend(inter_pair)
        return InteractingSystem(subsystem_list, inter_list)

    def convert_graph_to_pulse_lst(self, hilbert_space: InteractingSystem,
                                   **kwargs):
        """Obtain the pulse list and the max length of all pulses.
        The pulse list contain a list of [wrap_opt, pulse.create_pulse].
        wrap_opt is the drive operator on the whole Hilbert space.
        pulse.create_pulse gives function for computing the shape of the pulses.

        Args:
            hilbert_space: the hilbert space class
            kwargs: The keyword arguments be passed to module `PulseShape`

        """
        graph = copy.deepcopy(self)
        pulse_lst = []
        pulse_endtime = [0.]
        for node_name in self.sorted_nodes:
            node = graph.nodes[node_name]
            for k, v in dict(node).items():
                if str(k).startswith('pulse'):
                    opt_type = v.pop('operator_type')
                    opt = getattr(hilbert_space[node_name], opt_type)
                    pulse_type = v.pop('pulse_type')
                    delay = v.pop('delay', 0.)
                    # update global kwargs with graph kwargs
                    pulse_kwargs = copy.deepcopy(kwargs)
                    pulse_kwargs.update(v.pop('arguments', {}))
                    pulse = pulse_shape_dict[pulse_type](delay=delay,
                                                         name=_parse_pulse_name(
                                                             node_name, k,
                                                             pulse_type),
                                                         **pulse_kwargs)
                    wrap_opt = identity_wrap(opt, hilbert_space.subsystem_list)
                    pulse_lst.append([wrap_opt, pulse.create_pulse])
                    pulse_endtime.append(pulse.pulse_endtime)
                    crosstalk = v.pop('crosstalk', None)
                    if crosstalk is not None:
                        for xterm, coeff in crosstalk.items():
                            opt = getattr(hilbert_space[xterm], opt_type)
                            wrap_opt = identity_wrap(
                                opt, hilbert_space.subsystem_list, coeff=coeff)
                            pulse_lst.append([wrap_opt, pulse.create_pulse])
        return pulse_lst, jnp.max(jnp.array(pulse_endtime))

    def update_params(self,
                      params: dict,
                      share_params=True,
                      unify_coupling=False):
        """Update parameters to networkX Graph

        Args:
            params: new parameters dictionary
            share_params : If True, two sets of parameters will be shared
                according to the shared_param_mark attribute in the graph. This means
                that the shared parameters will remain the same during optimization.
                First, qubits with the same shared_param_mark will have their parameters
                shared. Second, the coupling strengths between qubits of
                type i and j are shared for all such pairs.
            unify_coupling: all coupling sharing the same parameters. Valid
                only if `share_params` is True.
        """
        # check arguments
        if unify_coupling:
            assert share_params is True
        if share_params:
            # load parameters dict
            type_dict = {}
            pulse_dict = {}
            edge_type_dict = {}
            for k, v in params.items():
                parse_k = k.split('_')
                # artificial atom
                if len(parse_k) == 1:
                    q = parse_k[0]
                    q_type = NodeView(self)[q].get('shared_param_mark', q)
                    type_dict.update({q_type: v})
                # pulse shape
                elif len(parse_k) == 3:
                    pulse_dict.update({k: v})
                # interaction
                elif len(parse_k) == 4:
                    k1, k2, q1, q2 = parse_k
                    coupling_name = '_'.join([k1, k2])
                    if '_'.join([q1, q2]) == 'all_unify':
                        try:
                            edge_type_dict[('all',
                                            'unify')].update({coupling_name: v})
                        except KeyError:
                            edge_type_dict.update({
                                ('all', 'unify'): {
                                    coupling_name: v
                                }
                            })
                    else:
                        edge_type = tuple(
                            sorted(
                                NodeView(self)[node].get(
                                    'shared_param_mark', node)
                                for node in [q1, q2]))
                        try:
                            edge_type_dict[edge_type].update({coupling_name: v})
                        except KeyError:
                            edge_type_dict.update(
                                {edge_type: {
                                    coupling_name: v
                                }})
            # update graph nodes parameters
            for node in NodeView(self):
                node_type = NodeView(self)[node].get('shared_param_mark', node)
                try:
                    NodeView(self)[node].update(type_dict[node_type])
                except KeyError:
                    warnings.warn(
                        f'The parameters dictionary does not contain nodes '
                        f'type {node_type}, keeping node {node} unchanged.')

            # update graph edges parameters
            for edge in EdgeView(self):
                q1, q2 = edge
                edge_type = tuple(
                    sorted([
                        NodeView(self)[node].get('shared_param_mark', node)
                        for node in [q1, q2]
                    ]))
                # update interaction
                try:
                    if unify_coupling:
                        EdgeView(self)[edge].update(
                            list(edge_type_dict.values())[0])
                    else:
                        EdgeView(self)[edge].update(edge_type_dict[edge_type])
                except KeyError and IndexError:
                    warnings.warn(
                        f'The parameters dictionary does not contain edges '
                        f'type {edge_type}, keeping edge {edge} unchanged.')
        else:
            pulse_dict = {}
            for k, v in params.items():
                parse_k = k.split('_')
                # artificial atom
                if len(parse_k) == 1:
                    q = parse_k[0]
                    NodeView(self)[q].update(v)
                # pulse shape
                elif len(parse_k) == 3:
                    pulse_dict.update({k: v})
                # interaction
                elif len(parse_k) == 4:
                    k1, k2, q1, q2 = parse_k
                    coupling_name = '_'.join([k1, k2])
                    EdgeView(self)[(q1, q2)].update({coupling_name: v})
        # update pulse shape parameters
        for node_name in self.sorted_nodes:
            node = NodeView(self)[node_name]
            for k, v in dict(node).items():
                if str(k).startswith('pulse'):
                    pulse_type = v['pulse_type']
                    pulse_info = pulse_dict.get(
                        _parse_pulse_name(node_name, k, pulse_type), None)
                    if pulse_info is not None:
                        v.update(pulse_info)

    def remove_pulse(self):
        """Remove pulse parameters from the graph"""
        for node_name in self.sorted_nodes:
            node = NodeView(self)[node_name]
            for k, v in dict(node).items():
                if str(k).startswith('pulse'):
                    # remove key and value pair
                    node.pop(k, None)



import numpy as np
import jax
import jax.numpy as jnp
from jax.experimental.ode import odeint

from supergrad.quantum_system import KronObj
from supergrad.time_evolution.ode import _parse_hamiltonian, ode_expm
from supergrad.utils.sharding import get_sharding


def sesolve(hamiltonian,
            psi0: jnp.ndarray,
            tlist: jnp.ndarray,
            args=None,
            options={},
            solver='ode_expm'):
    """
    Schrödinger equation evolution of a state vector or a set of state vectors
    for a given Hamiltonian.

    Evolve the state vector (``psi0``) using a given Hamiltonian (``H``).
    Alternatively evolve a set of states use vmap transformation.

    **Time-dependence formats**

    There are two major formats for specifying a time-dependent scalar:

    - Python function
    - array

    For function format, the function signature must be
    ``f(t: float, args: dict) -> complex``, for example

    .. code-block:: python

        def f1_t(t, args):
            return np.exp(-1j * t * args["w1"])

        def f2_t(t, args):
            return np.cos(t * args["w2"])

        H = _parse_sesolve([H0, [H1, f1_t], [H2, f2_t]], args={"w1":1., "w2":2.})

    For numpy array format, the array must be an 1d of dtype ``jnp.float64`` or
    ``jnp.complex128``.  A list of times (``jnp.float64``) at which the
    coeffients must be given as ``tlist``.  The coeffients array must have the
    same length as the tlist.  The times of the tlist do not need to be
    equidistant, but must be sorted.  By default, a linear interpolation
    will be used for the coefficient at time t. Examples of array-format usage are:

    .. code-block:: python

        tlist = np.logspace(-5, 0, 100)
        H = _parse_sesolve([H0, [H1, np.exp(-1j * tlist)], [H2, np.cos(2. * tlist)]],
                    tlist=tlist)

    Mixing time formats is allowed.  It is not possible to parse a single
    hamiltonian that contains different ``tlist`` values.

    Args:
        hamiltonian (list, array):
            System Hamiltonian as a :obj:`ndarray` , list of :obj:`ndarray` and
            coefficient. List format and options can be found in description.

        psi0 (array):
            Initial state vector (ket) or initial unitary operator ``psi0 = U``.
            Alternatively evolve a set of states use vmap transformation.

        tlist (list, array):
            List of times for :math:`t`.

        args (dict, optional):
            Dictionary of scope parameters for time-dependent Hamiltonians.

        options (dict, optional):
            Options for the ODE solver.
            `multi_device` (bool):
            `True` for enable multi-device parallelism evolution support,
            if many states are evolved simultaneously, the compiler can
            automatically batch the states to different devices.
            `diag_ops` (bool):
            It is only for `ode_expm`, if `True` the Hermitian Hamiltonian
            will be convert to diagonal form before exponentiation.

        solver (str, optional): select differentiable ODE solver.
            `odeint` for using Dormand-Prince ODE integration,
            `ode_expm` for using the matrix exponentiation at each time step.

    Returns: Values of the solution state vector at each time point in `tlist`,
        represented as an array with the same shape/structure as `psi0` except
        with a new leading axis of length `len(t)`. if a set of states be evolved
        simultaneously, the returns with another new leading axis of length `len(psi0)`.
    """
    diag_ops = options.pop('diag_ops', False)
    if diag_ops:
        assert solver == 'ode_expm'
    # convert options to hashable tuple
    options = tuple(options.items())

    if isinstance(hamiltonian, (KronObj, jnp.ndarray, list)):
        h_td = _parse_hamiltonian(hamiltonian, tlist, args, diag_ops)
    else:
        raise ValueError('Invalid Hamiltonian type.')

    if psi0.ndim == 2:
        return _sesolve(psi0, tlist, h_td, args, solver, options)
    elif psi0.ndim == 3:
        _simd_sesolve = jax.vmap(_sesolve,
                                 in_axes=(0, None, None, None, None, None))
        psi0 = jax.lax.with_sharding_constraint(psi0,
                                                get_sharding('p', None, None))
        return _simd_sesolve(psi0, tlist, h_td, args, solver, options)
    else:
        raise ValueError(
            '`psi0` is array or array of arrays representing the initial state')


def _sesolve(psi0, tlist, h_td, args, solver, options):
    """The private function of `sesolve`"""

    tlist = jnp.asarray(tlist)
    # Start evolution
    if solver == 'ode_expm':
        psi_evo = ode_expm(lambda psi, t: -1.0j * h_td(t, args=args), psi0,
                           tlist, **dict(options))
    elif solver == 'odeint':
        if isinstance(psi0, KronObj):
            psi0 = psi0.full()
        psi_evo = odeint(lambda psi, t: -1.0j * h_td(t, args=args) @ psi, psi0,
                         tlist, **dict(options))
    else:
        raise ValueError('Unknown ode solver.')

    return psi_evo


def sesolve_final_states_w_basis_trans(hamiltonian,
                                       psi_list,
                                       tlist,
                                       transform_matrix=None,
                                       **kwargs):
    """Compute the final states and applying basis transformation according to
    the transform_matrix.
    The time evolution is computed using sesolve in the product basis.

    Args:
        hamiltonian: `jax.numpy.ndarray`, list
            System Hamiltonian as a :obj:`~ndarray`, list of :obj:`ndarray` and
            coefficient. List format is the same as `qutip.sesolve`.

        psi_list: `jax.numpy.ndarray`
            Initial state vector (ket) or array of state vectors.

        tlist: array_like of float
            List of times for :math:`t`.

        transform_matrix: `jax.numpy.ndarray`
            The transform unitary from product basis to the desired basis.

        kwargs:
            Keyword arguments will be pass to `supergrad.time_evolution.sesolve`
    """
    # Find the corresponding basis in psi_list by the population
    # Note it should be convert to real type to compare
    if psi_list.ndim == 3:
        pop = np.abs(psi_list).real**2
        tuple_ar_ix = tuple(np.argmax(pop, axis=1).flatten())
    # evolution in the product basis with cycle transform
    if transform_matrix is not None:
        # Evolve in multi-qubit eigenbasis

        @jax.jit
        def _evolving_states(transform_matrix, psi_list):
            # Fix bug in gradient computation with multi-device
            # we use jit to tell the compiler transform_matrix must be provided
            # before the evolution
            psi_list = transform_matrix @ psi_list
            res = sesolve(hamiltonian, psi_list, tlist, **kwargs)
            if psi_list.ndim == 3:
                states = res[:, -1, :]
            else:
                states = res[-1, :]
            return jnp.conj(transform_matrix).T @ states

        states = _evolving_states(transform_matrix, psi_list)
    else:
        # Evolve in multi-qubit product basis
        res = sesolve(hamiltonian, psi_list, tlist, **kwargs)
        if psi_list.ndim == 3:
            states = res[:, -1, :]
        else:
            states = res[-1, :]
    if psi_list.ndim == 3:
        return states[:, tuple_ar_ix, 0].T
    else:
        return states



from copy import deepcopy
from collections import deque

import numpy as np
import networkx as nx

from supergrad.scgraph.graph import SCGraph
from supergrad.helper.compute_spectrum import Spectrum


class MPCFluxonium1D(SCGraph):
    """The general graph for 1d multipath coupling fluxonium chain. One could
    select a 3-frequency pattern and decide the number of qubits in the chain.
    The graph is then used to create a quantum processor, and one could attach
    control pulse to any qubit using class method.

    Args:
        n_qubit (int): number of qubits in the chain
        periodic (bool): whether the chain is periodic(the chain becomes a loop) or not
        seed (int, optional): random seed for the graph to add variation.
            Defaults to None.
    """

    def __init__(self, n_qubit: int = None, periodic=None, seed=None):
        super().__init__()

        # device parameters for frequency pattern
        fluxonium_type_1 = {
            "ec": 1.0 * 2 * np.pi,
            "ej": 4.0 * 2 * np.pi,
            "el": 0.9 * 2 * np.pi,
            "shared_param_mark": "grey",
            "phiext": np.pi,
            "system_type": "fluxonium",
        }

        fluxonium_type_2 = {
            "ec": 1.0 * 2 * np.pi,
            "ej": 4.0 * 2 * np.pi,
            "el": 1.0 * 2 * np.pi,
            "shared_param_mark": 'blue',
            "phiext": np.pi,
            "system_type": "fluxonium",
        }

        fluxonium_type_3 = {
            "ec": 1.0 * 2 * np.pi,
            "ej": 4.0 * 2 * np.pi,
            "el": 1.1 * 2 * np.pi,
            "shared_param_mark": "green",
            "phiext": np.pi,
            "system_type": "fluxonium",
        }

        mp_coupling = {
            "capacitive_coupling": {
                "strength": 20.0e-3 * 2 * np.pi
            },
            "inductive_coupling": {
                "strength": -1.0 * 2e-3 * 2 * np.pi
            },
        }

        # initialize graph
        if n_qubit is not None:
            temp_graph = nx.grid_graph([n_qubit], periodic=periodic)
            # adding attributes to nodes
            params = deque(
                [fluxonium_type_1, fluxonium_type_2, fluxonium_type_3])
            for i in range(n_qubit):
                temp_graph.nodes[i].update(params[i % 3])
            # relabel nodes
            label_mapping = dict((label, ''.join(['fm', str(label)]))
                                 for label in temp_graph.nodes)
            temp_graph = nx.relabel_nodes(temp_graph, label_mapping)
            # adding attributes to edges
            for edge in temp_graph.edges:
                temp_graph.edges[edge].update(mp_coupling)
            # save temp_graph
            self.add_nodes_from(temp_graph.nodes.data())
            self.add_edges_from(temp_graph.edges.data())

        if seed is not None:
            # add variance to el ec ej params
            self.add_lcj_params_variance_to_graph(multi_err=0.01, seed=seed)

    def compute_static_properties(self, list_drive_subsys, add_random,
                                  **kwargs):
        """Compute static properties of the model.

        Args:
            list_drive_subsys: list of driving subsystem, each element is a list
                of qubit names
            add_random: If true, will add random deviations to the device parameters
            kwargs: other parameters for the Spectrum class
        """
        # Compute the model once
        spec = Spectrum(self, truncated_dim=2, add_random=add_random, **kwargs)
        list_energy_nd, list_n_mat, list_phi_mat, transform_matrix = spec.get_model_eigen_basis(
            spec.all_params, self.sorted_nodes, list_drive_subsys)
        return list_energy_nd, list_n_mat, list_phi_mat, transform_matrix

    def compute_static_properties_minimal(self, list_drive_subsys, add_random,
                                          **kwargs):
        """Compute static properties of the minimal drive subsystem, as an
        approach to find quantum gate initial guesses.

        Args:
            list_drive_subsys: list of driving subsystem, each element is a list
                of qubit names
            add_random: If true, will add random deviations to the device parameters
            kwargs: other parameters for the Spectrum class
        """
        list_sub_energy_nd = []
        list_sub_n_mat = []
        list_sub_phi_mat = []
        for drive_subsys in list_drive_subsys:
            spec = Spectrum(self.subgraph(drive_subsys),
                            truncated_dim=2,
                            add_random=add_random,
                            **kwargs)
            list_energy_nd, list_n_mat, list_phi_mat, _ = spec.get_model_eigen_basis(
                spec.all_params, spec.graph.sorted_nodes, [drive_subsys])
            list_sub_energy_nd.append(list_energy_nd[0])
            list_sub_n_mat.append(list_n_mat[0])
            list_sub_phi_mat.append(list_phi_mat[0])

        return list_sub_energy_nd, list_sub_n_mat, list_sub_phi_mat, None

    def _add_crosstalk_pulse(self, name_drive, ar_crosstalk, pulse_dict):
        """Add crosstalk pulse based on the crosstalk matrix.

        Args:
            name_drive: the name of the drive qubit
            ar_crosstalk: the crosstalk matrix.  `ar[j,i]` means
                the pulse `A` applied on `i` qubit leads to `ar[j,i]*A` on `j`
                qubit.
            pulse_dict: the pulse dictionary
        """
        # Add full matrix
        dic2 = deepcopy(pulse_dict)
        dic2['crosstalk'] = {}
        for ix_name1, name in enumerate(self.sorted_nodes):
            # Keep only the wanted drive
            if name != name_drive:
                continue
            dic_c = {}
            for ix_name2, name2 in enumerate(self.sorted_nodes):
                v = ar_crosstalk[ix_name2, ix_name1]
                if ix_name1 != ix_name2 and v != 0:
                    dic_c[name2] = v
            # Only add when non-zero
            if (np.array(dic_c.values()) != 0).any():
                dic2["crosstalk"].update(dic_c)

        return dic2

    def create_cr_pulse(self,
                        ix_control_list: int,
                        ix_target_list: int,
                        tg_list: int,
                        add_random: bool,
                        ar_crosstalk=None,
                        pulse_type: str = "cos",
                        minimal_approach=True):
        """Creates CR pulses in the quantum processor.

        This function supports any multipath coupling model implemented.

        `index` is a model-dependent parameter depends on `self.sorted_nodes`

        Args:
            ix_control_list: list of the control qubit index, indicates where to
                apply the drive.
            ix_target_list: list of the target qubit index, indicates the target
                qubit that matched the drive frequency.
            tg_list: list of the gate time
            add_random: If true, will add random deviations to the device parameters
            ar_crosstalk: the crosstalk matrix.  `ar[j,i]` means
                the pulse `A` applied on `i` qubit leads to `ar[j,i]*A` on `j`
                qubit.
            pulse_type: the pulse type, by default it is "cos"
            minimal_approach: if True, will use the minimal subsystem approach to
                find the initial guess for the CNOT gate. Will not compute the
                transform matrix in this case.

        Returns:
            transform_matrix
        """
        # Get the initial guess from the Hamiltonian
        list_drive_subsys = []
        for control, target in zip(ix_control_list, ix_target_list):
            # check the target qubit in the neighbors of control qubit.
            if self.sorted_nodes[target] not in self.neighbors(
                    self.sorted_nodes[control]):
                raise ValueError(
                    f"Target qubit {target} is not a neighbor of control qubit {control}"
                )
            # the index of drive_subsys must match the intrinsic order
            list_drive_subsys.append(
                [self.sorted_nodes[ix] for ix in sorted([control, target])])
        if minimal_approach:
            list_energy_nd, list_n_mat, list_phi_mat, transform_matrix = \
                self.compute_static_properties_minimal(list_drive_subsys, add_random)
        else:
            list_energy_nd, list_n_mat, list_phi_mat, transform_matrix = \
                self.compute_static_properties(list_drive_subsys, add_random)

        for control, target, tg, drive_subsys, energy_nd, n_mat, phi_mat in zip(
                ix_control_list, ix_target_list, tg_list, list_drive_subsys,
                list_energy_nd, list_n_mat, list_phi_mat):

            s_start = (0, 0)
            name_drive = self.sorted_nodes[control]
            name_target = self.sorted_nodes[target]
            if name_drive == drive_subsys[0]:
                s_target = (0, 1)
                s_control = (1, 0)
            else:
                s_target = (1, 0)
                s_control = (0, 1)
            fd = abs(energy_nd[s_target] - energy_nd[s_start])
            f_control = abs(energy_nd[s_control] - energy_nd[s_start])
            detuning = abs(fd - f_control)
            # compute the effective coupling by n matrices and phi matrices
            ar_nd = np.array([
                n_drive[(*s_start, *s)]
                for n_drive, s in zip(n_mat, [s_control, s_target])
            ])
            ar_phi = np.array([
                phi_drive[(*s_start, *s)]
                for phi_drive, s in zip(phi_mat, [s_control, s_target])
            ])
            jc = self.get_edge_data(
                name_drive, name_target)['capacitive_coupling']['strength']
            jl = self.get_edge_data(
                name_drive, name_target)['inductive_coupling']['strength']
            j_eff = np.abs(np.prod(ar_nd) * jc + np.prod(ar_phi) * jl)
            tau_eps_drive = np.pi / 2.0 * detuning / j_eff

            cr_pulse = {
                "amp": tau_eps_drive / tg,
                "omega_d": fd,
                "phase": 0.0,
                "length": float(tg),
                "pulse_type": pulse_type,
                "operator_type": "phi_operator",
                "delay": 0.0,
            }

            if ar_crosstalk is None:
                dic1 = deepcopy(cr_pulse)
                self.add_node(name_drive, pulsecr=dic1)
            else:
                dic2 = self._add_crosstalk_pulse(name_drive, ar_crosstalk,
                                                 cr_pulse)
                self.add_node(name_drive, pulsecr=dic2)
        return transform_matrix

    def create_single_qubit_pulse(self,
                                  ix_qubit_list: int,
                                  tg_list: int,
                                  add_random: bool,
                                  ar_crosstalk=None,
                                  factor=0.5,
                                  pulse_type: str = "cos",
                                  minimal_approach=False):
        """Creates single qubit pulses in the quantum processor.

        This function supports any multipath coupling model implemented.

        `index` is a model-dependent parameter depends on `self.sorted_nodes`

        Args:
            ix_qubit_list: target qubit index. `ix_pair` is a model-dependent
            parameter, is the qubit index, indicates where to apply the drive.
            tg_list: the gate time
            add_random: If true, will add random deviations to the device parameters
            ar_crosstalk: the crosstalk matrix.  `ar[j,i]` means
                the pulse `A` applied on `i` qubit leads to `ar[j,i]*A` on `j`
                qubit.
            factor: the amplitude factor, driving a X gate by default.
            pulse_type: the pulse type, by default it is "cos"
            minimal_approach: if True, will use the minimal subsystem approach to
                find the initial guess for the single qubit gate. Will not compute
                the transform matrix in this case.

        Returns:
            transform_matrix
        """
        # Get the initial guess from the Hamiltonian
        list_drive_subsys = [[self.sorted_nodes[ix]] for ix in ix_qubit_list]
        if minimal_approach:
            list_energy_nd, list_n_mat, list_phi_mat, transform_matrix = \
                self.compute_static_properties_minimal(list_drive_subsys, add_random)
        else:
            list_energy_nd, list_n_mat, list_phi_mat, transform_matrix = \
                self.compute_static_properties(list_drive_subsys, add_random)

        for ix_qubit, tg, energy_nd, n_mat, phi_mat in zip(
                ix_qubit_list, tg_list, list_energy_nd, list_n_mat,
                list_phi_mat):
            name_drive = self.sorted_nodes[ix_qubit]
            s_start = (0,)
            s_end = (1,)

            # nd = abs(n_mat[0][(*s_start, *s_end)])
            phid = abs(phi_mat[0][(*s_start, *s_end)])
            fd = abs(energy_nd[s_start] - energy_nd[s_end])

            dic_pulse = {
                'amp': factor * 2 / tg / phid * 2 * np.pi,
                'omega_d': fd,
                "phase": 0.0,
                "length": float(tg),
                "pulse_type": pulse_type,
                "operator_type": "phi_operator",
                "delay": 0.0,
            }

            if ar_crosstalk is None:
                # update amplitude
                new_pulse = deepcopy(dic_pulse)
                self.add_node(name_drive, pulsesq=new_pulse)
            else:
                dic2 = self._add_crosstalk_pulse(name_drive, ar_crosstalk,
                                                 dic_pulse)
                self.add_node(ix_qubit, pulsesq=dic2)
        return transform_matrix



"""The Lindblad Object (LindbladObj) class is designed to represent superoperator
in Kronecker product format. It's particularly useful for representing sparse total systems'
superoperator in a form convenient for computing Einstein summation.
"""
import numbers
from math import sqrt
import copy

import numpy as np
import jax
from jax.tree_util import register_pytree_node_class
import jax.numpy as jnp
import jax.scipy.linalg as jaxLA
import opt_einsum as oe

from supergrad.utils.utility import tensor
from supergrad.quantum_system.kronobj import KronObj


@register_pytree_node_class
class LindbladObj(object):
    """The Lindblad Object (LindbladObj) class is designed to represent
    superoperator in Kronecker product format. It's particularly useful for
    representing sparse total systems' superoperator in a form convenient for
    computing Einstein summation.

    We express the Lindblad equation for a tensor-4 superoperator, the Hamiltonian
    part and Lindblad generator part are expressed as the tensor product of two
    operators. The local Hamiltonian and corresponding location should be
    specified in the `inpt` and `locs` argument.

    Args:
        inpt (list): list of list of array
            `inpt` is a nested list represents a tensor product, while the first
            list corresponds to the left operand and the second list corresponds
            to right operand. Each element of `operand` corresponds to
            the local Hamiltonian of the corresponding subsystem tagged in the `locs`.
        dims (list): list of int
            Dimensions of subsystems in the composited Hilbert Space.
        locs (list): list of int
            Location information of corresponding local Hamiltonian. `locs` is a
            nested list represents a tensor product, while the first list corresponds
            to the left operand and the second list corresponds to right operand.
            For example, the first element of left operand is the index of the
            first subsystem in `self.dims`, and so on.
        diag_unitary (array):
            Unitary of diagonal-unitary format.
        _nested_inpt (bool):
            For internal using only.
    """

    def __init__(self,
                 inpt=None,
                 dims=None,
                 locs=None,
                 _nested_inpt=False,
                 diag_unitary=None) -> None:
        if _nested_inpt:
            # identify diag_unitary method
            if diag_unitary is None:
                self.diag_unitary = [[None, None]] * len(inpt)
            else:
                self.diag_unitary = list(diag_unitary)
            # identify inpt location
            if locs is None:
                self.locs = [[None, None]] * len(inpt)
            else:
                self.locs = list(locs)
            self._data = []
            self.dims = dims
            # check the data formation
            for liou_region in inpt:
                self._data.append([ele for ele in liou_region])
        elif inpt is None:
            # idle constructor
            self.locs = []
            self.diag_unitary = []
            self._data = []
            self.dims = None
        else:
            # identify diagonal-unitary representation
            if diag_unitary is None or None in diag_unitary:
                self.diag_unitary = [[None, None]]
            else:
                # for ul_region in inpt:
                #     if ul_region and ul_region[0].ndim == 1:
                #         # convert eigenenergies to diagonal matrix
                #         ul_region = [jnp.diag(ul_region[0])]
                self.diag_unitary = [diag_unitary]  # nested list
            # identify localized method
            if locs is None:
                self.locs = [[None, None]]
            else:
                self.locs = [locs]
            self._data = [[ele for ele in inpt]]  # nested
            if dims is None:
                raise ValueError('No Hilbertspace dimensions. ')
            self.dims = dims

    def tree_flatten(self):
        """Specifies a flattening recipe.

        Returns:
            a pair of an iterable with the children to be flattened recursively,
            and some opaque auxiliary data to pass back to the unflattening recipe.
            The auxiliary data is stored in the treedef for use during unflattening.
            The auxiliary data could be used, e.g., for dictionary keys.
        """
        return ((self.data, self.diag_unitary), (self.dims, self.locs))

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Specifies an unflattening recipe.

        Args:
            aux_data: the opaque data that was specified during flattening of the
            current treedef.
            children: the unflattened children

        Returns:
            a re-constructed object of the registered type, using the specified
            children and auxiliary data.
        """
        dims, locs = aux_data
        data, diag_unitary = children
        return cls(data,
                   dims,
                   locs,
                   _nested_inpt=True,
                   diag_unitary=diag_unitary)

    def get_data(self):
        return self._data

    def set_data(self, data):
        if not isinstance(data[0], list):
            raise TypeError('LindbladObj data must be in a nested list.')
        else:
            self._data = data

    data = property(get_data, set_data)

    def _build_sub_lindbladobj(self, idx):
        """Return a subset of `LindbladObj` at selected `idx`."""
        return LindbladObj(self.data[idx],
                           dims=self.dims,
                           locs=self.locs[idx],
                           diag_unitary=self.diag_unitary[idx])

    @property
    def shape(self):
        """The shape of unitary"""
        res = [np.prod(self.dims), np.prod(self.dims)] * 2
        return res

    @property
    def loc_status(self):
        """The local format status of subsystems."""
        return [[ul_loc_info is not None
                 for ul_loc_info in liou_loc_info]
                for liou_loc_info in self.locs]

    @property
    def diag_status(self):
        """The diagonal-unitary format status of subsystems."""
        return [[ul_eig_vec is not None
                 for ul_eig_vec in liou_diag_info]
                for liou_diag_info in self.diag_unitary]

    def __str__(self):
        s = ""
        shape = self.shape
        s += ("Lindblad object: " + "dims = " + str(self.dims) + ", shape = " +
              str(shape) + ", diag_status = " + str(self.diag_status) +
              ", location_info = " + str(self.locs) + "\n")
        s += "LindbladObj data =\n"
        s += str(self.data)

        return s

    def __repr__(self):
        # give complete information on LindbladObj without print statement in
        # command-line we cant realistically serialize a LindbladObj into a string,
        # so we simply return the informal __str__ representation instead.)
        return self.__str__()

    def _downcast_diagonal_unitary(self):
        """Convert the diagonal-unitary format to order-4 tensor."""
        data = []
        for liou_region, liou_diag_info in zip(self.data, self.diag_unitary):
            if None not in liou_diag_info:  # downcasting
                liou_list = []
                if all(liou_diag_info):
                    raise ValueError('Unsupported diagonal-unitary info.')
                for ul_region, ul_eig_vec in zip(liou_region, liou_diag_info):
                    if ul_region:
                        liou_list.append([
                            ul_eig_vec @ ul_region[0] @ jnp.conj(ul_eig_vec).T
                        ])
                    else:
                        liou_list.append([])
                data.append(liou_list)
            else:
                data.append(liou_region)
        return LindbladObj(data,
                           _nested_inpt=True,
                           dims=self.dims,
                           locs=self.locs)

    def _downcast_local_hamiltonian(self):
        """Convert the local format to the order-4 tensor."""
        # firstly downcast diag unitary
        data = []
        for liou_region, liou_local in zip(self.data, self.locs):
            if None not in liou_local:  # downcasting to order-4 tensor
                liou_list = []
                for ul_region, ul_local in zip(liou_region, liou_local):
                    if ul_local is None:
                        liou_list.append(ul_region[0])
                    elif ul_local:  # downcasting to matrix
                        # construct identity matrix list
                        mat_list = [jnp.eye(dim) for dim in self.dims]
                        for mat, local in zip(ul_region, ul_local):
                            mat_list[local] = mat
                        # calculate tensor product
                        liou_list.append(tensor(*mat_list))
                    else:  # []
                        liou_list.append(jnp.eye(np.prod(self.dims)))
                data.append([tensor(*liou_list).reshape(self.shape)])
            else:
                data.append(liou_local)

        return LindbladObj(data, _nested_inpt=True, dims=self.dims)

    def full(self):
        """Dense order-4 tensor from the result of Kronecker product."""
        new_self = self
        # downcast the diag unitary
        if any(self.diag_status):
            new_self = new_self._downcast_diagonal_unitary()
        # downcast the location representation
        if any(self.loc_status):
            new_self = new_self._downcast_local_hamiltonian()
        ts_list = [ts[0] for ts in new_self.data]
        return sum(ts_list)

    def __iter__(self):
        """iteration through nested list."""
        sublist = [
            self._build_sub_lindbladobj(ind) for ind, _ in enumerate(self.data)
        ]
        return iter(sublist)

    def __getitem__(self, key):

        if isinstance(key, int):
            return self._build_sub_lindbladobj(key)
        raise KeyError(f"Unrecognized key: {key}. Key must be an integer index")

    def add_liouvillian(self, ham: KronObj):
        """Convert Hamiltonian to Liouvillian.

        Args:
            ham (`KronObj`) : Hamiltonian
        """
        # Liouvillian is a 4-order tensor, but each index share the same dimension
        if self.dims is None:
            self.dims = ham.dims
        else:
            assert self.dims == ham.dims
        for aham in ham:
            if aham.diag_unitary[0] is None:
                part_diag_info = None
            else:
                part_diag_info = 0
            self.data += [[[], (-1.0j * aham).data[0]]]
            self.locs += [[[], aham.locs[0]]]
            self.diag_unitary += [[part_diag_info, aham.diag_unitary[0]]]

            self.data += [[(1.0j * aham).transpose().data[0], []]]
            self.locs += [[aham.locs[0], []]]
            self.diag_unitary += [[aham.diag_unitary[0], part_diag_info]]

    def add_lindblad_operator(self, a: KronObj, b: KronObj = None, chi=None):
        """Lindblad operator for a single pair of collapse operators
        (a, b), or for a single collapse operator (a) when b is not specified:

        .. math::

            \\mathcal{D}[a,b]\\rho = a \\rho b^\\dagger -
            \\frac{1}{2}a^\\dagger b\\rho - \\frac{1}{2}\\rho a^\\dagger b

        Args:
            a: KronObj
                Left part of collapse operator.
            b: KronObj
                Right part of collapse operator. If not specified, b defaults to a.
            chi: float
        """
        if self.dims is None:
            self.dims = a.dims
        else:
            assert self.dims == a.dims
        if b is None:
            temp_b = a
        else:
            assert b.dims == a.dims
            temp_b = b

        for aa in a:
            for bb in temp_b:
                if chi:
                    dissi0_data = [
                        bb.conjugate().data[0],
                        (jnp.exp(1.0j * chi) * aa).data[0]
                    ]
                else:
                    dissi0_data = [bb.conjugate().data[0], aa.data[0]]
                self.data += [dissi0_data]
                self.locs += [[bb.locs[0], aa.locs[0]]]
                self.diag_unitary += [[aa.diag_unitary[0], bb.diag_unitary[0]]]
                if b is None and len(a.data) == 1 and len(
                        a.data[0]) == 1 and all(a.diag_status):
                    # The product of two Kronecker products yields another Kronecker product
                    # output keeps the diag representation
                    # only support regular dissipator collapse_b = collapse_a
                    # note here kron_a = kron_b.dag()
                    # new_diag = -0.5 * jnp.diag(
                    #     a.data[0][0]).conjugate() * jnp.diag(a.data[0][0])
                    new_diag = -0.5 * jnp.conjugate(a.data[0][0]) * a.data[0][0]
                    ad_b = KronObj([new_diag],
                                   a.dims,
                                   locs=a.locs[0],
                                   diag_unitary=a.diag_unitary[0])
                else:
                    ad_b = -0.5 * aa.dag() @ bb
                if ad_b.diag_unitary[0] is None:
                    part_diag_info = None
                else:
                    part_diag_info = 0
                self.data += [[[], ad_b.data[0]]]
                self.locs += [[[], ad_b.locs[0]]]
                self.diag_unitary += [[part_diag_info, ad_b.diag_unitary[0]]]

                self.data += [[ad_b.transpose().data[0], []]]
                self.locs += [[ad_b.locs[0], []]]
                self.diag_unitary += [[ad_b.diag_unitary[0], part_diag_info]]

    def _diagonalize_super_operator(self):
        """Diagonalize super operator."""
        for idx, liou_region in enumerate(self.data):
            if all(liou_region):
                union_ul_local = sorted(
                    set(self.locs[idx][0] + self.locs[idx][1]))
                ul_list = []
                for ul_region, ul_local in zip(liou_region, self.locs[idx]):
                    mat_list = [
                        jnp.eye(self.dims[loc]) for loc in union_ul_local
                    ]
                    for mat, loc in zip(ul_region, ul_local):
                        loc_idx = union_ul_local.index(loc)
                        mat_list[loc_idx] = mat
                    ul_list.append(tensor(*mat_list))
                # only calculate matrix exponentiation with the terms in `local_info`
                obj = tensor(*ul_list[::-1])
                # we cannot diagonalize a non-Hermite matrix
                # TODO: add jit support
                if not jnp.allclose(jnp.conj(obj).T, obj):
                    continue
                eig, eigv = jaxLA.eigh(obj)
                # diagonalize the super operator
                tensor_dim = np.prod([self.dims[loc] for loc in union_ul_local])
                # self.data[idx] = [jnp.diag(eig).reshape(*[tensor_dim] * 4)]
                self.data[idx] = [eig.reshape(*[tensor_dim] * 2)]
                self.diag_unitary[idx] = eigv.reshape(*[tensor_dim] * 4)
                self.locs[idx] = [union_ul_local, union_ul_local]
            else:
                mat = tensor(*(liou_region[0] + liou_region[1]))
                # we cannot diagonalize a non-Hermite matrix
                # TODO: add jit support
                if not jnp.allclose(jnp.conj(mat).T, mat):
                    continue
                eig, eigv = jaxLA.eigh(mat)
                if liou_region[0]:
                    self.data[idx] = [[eig], []]
                    self.diag_unitary[idx] = [eigv, 0]
                else:
                    self.data[idx] = [[], [eig]]
                    self.diag_unitary[idx] = [0, eigv]

    def __add__(self, other):
        """Addition with LindbladObj on left. The trivial addition is defined as
        concatenating the list.
        """
        if other == 0:  # identity of addition operator
            return self
        else:  # case for matching KronObj
            return LindbladObj(self.data + other.data,
                               self.dims,
                               _nested_inpt=True,
                               locs=self.locs + other.locs,
                               diag_unitary=self.diag_unitary +
                               other.diag_unitary)

    def __radd__(self, other):
        """Addition with KronObj on right. Note the addition satisfies the
        commutative.
        """
        return self + other

    def __sub__(self, other):
        """Subtraction with KronObj on left."""
        return self + (-1.0 * other)

    def __rsub__(self, other):
        """Subtraction with KronObj on right."""
        return (-1.0 * self) + other

    def __mul__(self, other):
        """Multiplication with KronObj on left, the operator only acts on first
        term in the nested list. Note the scale multiplication satisfies the
        commutative.
        """
        if isinstance(other, (numbers.Number, jax.Array)):
            data = []
            for liou_region in self.data:
                if isinstance(liou_region[0], list):
                    if liou_region[0]:
                        # liou_region = [[3], [2]] or [[3,2], []]
                        # multiply scale number to first term
                        array_0 = liou_region[0][0] * other
                        temp_data = copy.deepcopy(liou_region)
                        temp_data[0][0] = array_0
                        data.append(temp_data)
                    elif liou_region[1]:
                        # liou_region = [[], [3,2]]
                        array_0 = liou_region[1][0] * other
                        temp_data = copy.deepcopy(liou_region)
                        temp_data[1][0] = array_0
                        data.append(temp_data)
                else:
                    # order-4 tensor be used as Lindblad
                    data.append([liou_region[0] * other])
            return LindbladObj(data,
                               self.dims,
                               _nested_inpt=True,
                               locs=self.locs,
                               diag_unitary=self.diag_unitary)
        else:
            raise NotImplementedError(f'Do not support instance {type(other)}.')

    def __rmul__(self, other):
        """Multiplication with KronObj on right"""
        return self * other

    def __matmul__(self, other):
        """Matrix multiplication with LindbladObj on left. The high efficient
        order-4 tensor <-> density matrix multiplication method working when `other` has matched shape.
        """
        if isinstance(other, (np.ndarray, jax.Array)):
            res = []
            # implement location Einstein summation
            for sub_lindbladobj in self:
                # construct subscripts and operands
                lindblad_op = []
                lindblad_dim = []
                lindblad_subscript = []
                if isinstance(sub_lindbladobj.data[0][0], list):
                    if all(sub_lindbladobj.locs[0]):
                        # both upper region and lower region have operator
                        union_loc_info = sorted(
                            set(sub_lindbladobj.locs[0][0] +
                                sub_lindbladobj.locs[0][1]))
                        for idx, ul_region, ul_loc in zip(
                                range(2), sub_lindbladobj.data[0],
                                sub_lindbladobj.locs[0]):
                            mat_list = [
                                jnp.eye(self.dims[loc])
                                for loc in union_loc_info
                            ]
                            for mat, loc in zip(ul_region, ul_loc):
                                mat_list[union_loc_info.index(loc)] = mat
                            lindblad_op.append(tensor(*mat_list))
                            lindblad_dim += [
                                sub_lindbladobj.dims[loc]
                                for loc in union_loc_info
                            ] * 2
                            # subscript start from 1
                            # for Lindblad L_{ijkl}
                            # | subscript_l | -len(dims) subscript_j | 0 | subscript_i len(dims) | subscript_k |
                            temp_subscript = [
                                len(sub_lindbladobj.dims) * idx + 1 + loc
                                for loc in union_loc_info
                            ]
                            lindblad_subscript += temp_subscript
                            lindblad_subscript += [
                                -subscript for subscript in temp_subscript
                            ]
                    else:
                        for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                            if ul_loc:
                                lindblad_op.append(
                                    tensor(*sub_lindbladobj.data[0][idx]))
                                lindblad_dim += [
                                    sub_lindbladobj.dims[loc] for loc in ul_loc
                                ] * 2
                                # subscript start from 1
                                # for Lindblad L_{ijkl}
                                # | subscript_l | -len(dims) subscript_j | 0 | subscript_i len(dims) | subscript_k |
                                temp_subscript = [
                                    len(sub_lindbladobj.dims) * idx + 1 + loc
                                    for loc in ul_loc
                                ]
                                lindblad_subscript += temp_subscript
                                lindblad_subscript += [
                                    -subscript for subscript in temp_subscript
                                ]
                            else:  # []
                                other_ul_loc = sub_lindbladobj.locs[0][1 - idx]
                                temp_lindblad_dim = [
                                    sub_lindbladobj.dims[loc]
                                    for loc in other_ul_loc
                                ]
                                lindblad_op.append(
                                    jnp.eye(np.prod(temp_lindblad_dim)))
                                lindblad_dim += temp_lindblad_dim * 2
                                temp_subscript = [
                                    len(sub_lindbladobj.dims) * idx + 1 + loc
                                    for loc in other_ul_loc
                                ]
                                lindblad_subscript += temp_subscript
                                lindblad_subscript += [
                                    -subscript for subscript in temp_subscript
                                ]
                    lindblad_op = tensor(
                        *lindblad_op[::-1]).reshape(lindblad_dim)
                else:
                    # use order-4 tensor as Lindblad
                    for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                        lindblad_dim += [
                            sub_lindbladobj.dims[loc] for loc in ul_loc
                        ] * 2
                        temp_subscript = [
                            len(sub_lindbladobj.dims) * idx + 1 + loc
                            for loc in ul_loc
                        ]
                        lindblad_subscript += temp_subscript
                        lindblad_subscript += [
                            -subscript for subscript in temp_subscript
                        ]
                    lindblad_op = sub_lindbladobj.data[0][0].reshape(
                        lindblad_dim)
                # density matrix in the dense format
                density_op = other.reshape(sub_lindbladobj.dims * 2)
                # for denisty matrix rho_{kl}
                # | subscript_l | -len(dims) None | 0 | None len(dims) | subscript_k |
                density_subscript = []
                temp_subscript = list(
                    range(
                        len(sub_lindbladobj.dims) + 1,
                        len(sub_lindbladobj.dims) * 2 + 1))
                density_subscript += temp_subscript
                density_subscript += [
                    -subscript for subscript in temp_subscript
                ]
                res_subscript = density_subscript.copy()
                for subscript in lindblad_subscript:
                    if abs(subscript) > len(sub_lindbladobj.dims):
                        idx = res_subscript.index(subscript)
                        res_subscript[idx] -= len(
                            sub_lindbladobj.dims) * np.sign(res_subscript[idx])

                new_density = jnp.einsum(lindblad_op, lindblad_subscript,
                                         density_op, density_subscript,
                                         res_subscript).reshape(other.shape)
                res.append(new_density)
            return sum(res)
        else:
            raise NotImplementedError(f'Do not support instance {type(other)}.')

    def compute_contraction_path(self, lindblad_list=None, trotter_order=None):
        """Compute tensor network contraction path.
        Please provide either `lindblad_list` or `trotter_order`. If `lindblad_list` is None,
        it will be generated by `trotter_order`.

        Args:
            lindblad_list: the list of local operator
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
        """
        assert lindblad_list is not None or trotter_order is not None
        if lindblad_list is None:
            lindblad_list = self._trotter_decomposition(trotter_order)
        # initialize subscripts by state
        num_indices = len(self.dims)
        max_indices_per_layer = len(self.dims)**2
        num_subscript = np.zeros((2, len(self.dims)), dtype=int)
        rho_subscript = ''.join([
            oe.get_symbol(i + max_indices_per_layer * num)
            for i, num in enumerate(num_subscript.flatten())
        ])
        oe_subscripts = []
        oe_shapes = []
        # construct tensor network
        oe_subscripts.append(rho_subscript)
        oe_shapes.append(list(self.dims) * 2)
        for sub_lindbladobj in lindblad_list[::-1]:
            # construct subscripts and operands
            base_subscript = []
            new_subscript = []
            lindblad_shape = []
            if isinstance(sub_lindbladobj.data[0][0], list):
                for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                    if ul_loc:
                        lindblad_shape += [
                            sub_lindbladobj.dims[loc] for loc in ul_loc
                        ] * 2
                        for loc in ul_loc:
                            base_subscript.append(
                                oe.get_symbol(loc + num_indices * idx +
                                              max_indices_per_layer *
                                              num_subscript[idx, loc]))
                            num_subscript[idx, loc] += 1
                            new_subscript.append(
                                oe.get_symbol(loc + num_indices * idx +
                                              max_indices_per_layer *
                                              num_subscript[idx, loc]))
                    else:  # []
                        other_ul_loc = sub_lindbladobj.locs[0][1 - idx]
                        temp_lindblad_dim = [
                            sub_lindbladobj.dims[loc] for loc in other_ul_loc
                        ]
                        lindblad_shape += temp_lindblad_dim * 2
                        for loc in other_ul_loc:
                            base_subscript.append(
                                oe.get_symbol(loc + num_indices * idx +
                                              max_indices_per_layer *
                                              num_subscript[idx, loc]))
                            num_subscript[idx, loc] += 1
                            new_subscript.append(
                                oe.get_symbol(loc + num_indices * idx +
                                              max_indices_per_layer *
                                              num_subscript[idx, loc]))
            else:
                # use order-4 tensor as Lindblad
                for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                    lindblad_shape += [
                        sub_lindbladobj.dims[loc] for loc in ul_loc
                    ] * 2
                    for loc in ul_loc:
                        base_subscript.append(
                            oe.get_symbol(loc + num_indices * idx +
                                          max_indices_per_layer *
                                          num_subscript[idx, loc]))
                        num_subscript[idx, loc] += 1
                        new_subscript.append(
                            oe.get_symbol(loc + num_indices * idx +
                                          max_indices_per_layer *
                                          num_subscript[idx, loc]))
            oe_subscripts.append(''.join(new_subscript + base_subscript))
            oe_shapes.append(lindblad_shape)
        rho_output = ''.join([
            oe.get_symbol(i + max_indices_per_layer * num)
            for i, num in enumerate(num_subscript.flatten())
        ])
        eq = ','.join(oe_subscripts) + '->' + rho_output
        expr = oe.contract_expression(eq, *oe_shapes, optimize='auto-hq')
        return expr

    def _trotter_decomposition(self, trotter_order):
        """Generate Suzuki-Trotter decomposition sequence.

        Args:
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
        """

        def _loop(trotter_coeff, reverse=False):
            """Generate the matrix multiplication sequence.

            Args:
                trotter_coeff (float): the trotter coefficient
                reverse (bool): whether reverse the sequence or not.
            """
            lindblad_list = []
            if reverse:
                scan_list = zip(reversed(self.data),
                                reversed(self.diag_unitary),
                                reversed(self.locs))
            else:
                scan_list = zip(self.data, self.diag_unitary, self.locs)
            for liou_region, liou_diag, liou_local in scan_list:
                if all(liou_local):
                    # compute tensor product of upper_region and lower_region
                    union_ul_local = sorted(set(liou_local[0] + liou_local[1]))
                    if None not in liou_diag:  # cast expm down to exp
                        # diag info will be a order-4 tensor
                        eig_tensor = liou_diag
                        tensor_dim = np.prod(
                            [self.dims[loc] for loc in union_ul_local])
                        eig_vec = eig_tensor.reshape(*[tensor_dim**2] * 2)
                        # fix bug about zip function
                        # lam = jnp.exp(trotter_coeff * jnp.diag(
                        #     liou_region[0].reshape(*[tensor_dim**2] * 2)))
                        lam = jnp.exp(trotter_coeff * liou_region[0].flatten())
                        liou_local = [union_ul_local, union_ul_local]
                        tensor_expm = eig_vec * lam @ jnp.conj(eig_vec).T
                        inpt = [tensor_expm.reshape(*([
                            tensor_dim,
                        ] * 4))]
                    else:
                        ul_list = []
                        for ul_region, ul_local in zip(liou_region, liou_local):
                            mat_list = [
                                jnp.eye(self.dims[loc])
                                for loc in union_ul_local
                            ]
                            for mat, loc in zip(ul_region, ul_local):
                                loc_idx = union_ul_local.index(loc)
                                mat_list[loc_idx] = mat
                            ul_list.append(tensor(*mat_list))
                        # only calculate matrix exponentiation with the terms in `local_info`
                        obj = tensor(*ul_list[::-1]) * trotter_coeff
                        liou_local = [union_ul_local, union_ul_local]
                        tensor_dim = np.prod(
                            [self.dims[loc] for loc in union_ul_local])
                        inpt = [jaxLA.expm(obj).reshape(*([
                            tensor_dim,
                        ] * 4))]
                    lindblad_list.append(
                        LindbladObj(inpt, self.dims, liou_local))
                else:
                    # only one of region has non-trivial matrix
                    mat_list = liou_region[0] + liou_region[1]
                    if None not in liou_diag:  # cast expm down to exp
                        eig_vec = liou_diag[0] + liou_diag[1]
                        lam = jnp.exp(trotter_coeff * mat_list[0])
                        if liou_region[0]:
                            inpt = [[eig_vec * lam @ jnp.conj(eig_vec).T], []]
                        else:
                            inpt = [[], [eig_vec * lam @ jnp.conj(eig_vec).T]]
                    else:
                        obj = tensor(*mat_list) * trotter_coeff
                        if liou_region[0]:
                            inpt = [[jaxLA.expm(obj)], []]
                        else:
                            inpt = [[], [jaxLA.expm(obj)]]

                    lindblad_list.append(
                        LindbladObj(inpt, self.dims, liou_local))
                # note the `expm(obj)` cannot decomposition back to tensor product,
                # the `LindbladObj.full()` method does not support the form,
                # so `LindbladObj([expm(obj)])` only work in tensor-density multiplication
            return lindblad_list

        def _second_order_expm(p=1.0):
            """Generate the matrix multiplication sequence for second order
            trotter decomposition.
            """
            # normal loop
            lindblad_list = _loop(p / 2., reverse=False)
            # reversed loop
            lindblad_list.extend(_loop(p / 2., reverse=True))
            return lindblad_list

        if trotter_order == 1:
            lindblad_list = _loop(1.)
        elif trotter_order == 2:
            lindblad_list = _second_order_expm()
        elif trotter_order == 4:
            # the simplest real decomposition of 4th order STD
            p = 1 / (2 - 2**(1 / 3))
            lindblad_list = _second_order_expm(p)
            lindblad_list.extend(_second_order_expm(1 - 2 * p))
            lindblad_list.extend(_second_order_expm(p))
        elif trotter_order == 4j:
            # the complex 4th STD
            p = (3 - 1j * sqrt(3)) / 6
            p_conj = (3 + 1j * sqrt(3)) / 6
            lindblad_list = _second_order_expm(p)
            lindblad_list.extend(_second_order_expm(p_conj))
        else:
            raise ValueError(f'Unsupported trotter order {trotter_order}.')

        return lindblad_list

    def expm(self, right_density=None, trotter_order=None, tn_expr=None):
        """Calculate matrix exponentiation of `LindbladObj`. Combine matrix exponentiation
        with matrix-vector product if `right_vec` is not `None`.

        Args:
            right_vec (bool): matrix-vector product with the `right_vec` on the right
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
            tn_expr: Tensor network contract expression.
                if `tn_expr` is None, generate contract expression when `expm()`
                be called each time.

        """
        if right_density is None or trotter_order not in [1, 2, 4, 4j]:
            # not implement trotter approximate
            res = jaxLA.expm(self.full().reshape(np.prod(self.shape[:2]),
                                                 np.prod(self.shape[2:])))
            if right_density is not None:
                return (res @ right_density.ravel('F')).reshape(
                    np.prod(self.dims), np.prod(self.dims)).T
            else:
                return res
        else:
            lindblad_list = self._trotter_decomposition(trotter_order)
            if tn_expr is None:
                tn_expr = self.compute_contraction_path(lindblad_list)
            oe_arrays = []
            oe_arrays.append(right_density.reshape(*self.dims * 2))
            for sub_lindbladobj in lindblad_list[::-1]:
                # construct subscripts and operands
                lindblad_shape = []
                lindblad_op = []
                if isinstance(sub_lindbladobj.data[0][0], list):
                    for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                        if ul_loc:
                            lindblad_op.append(
                                tensor(*sub_lindbladobj.data[0][idx]))
                            lindblad_shape += [
                                sub_lindbladobj.dims[loc] for loc in ul_loc
                            ] * 2

                        else:  # []
                            other_ul_loc = sub_lindbladobj.locs[0][1 - idx]
                            temp_lindblad_dim = [
                                sub_lindbladobj.dims[loc]
                                for loc in other_ul_loc
                            ]
                            lindblad_op.append(
                                jnp.eye(np.prod(temp_lindblad_dim)))
                            lindblad_shape += temp_lindblad_dim * 2
                    lindblad_op = tensor(
                        *lindblad_op[::-1]).reshape(lindblad_shape)
                else:
                    # use order-4 tensor as Lindblad
                    for idx, ul_loc in enumerate(sub_lindbladobj.locs[0]):
                        lindblad_shape += [
                            sub_lindbladobj.dims[loc] for loc in ul_loc
                        ] * 2
                    lindblad_op = sub_lindbladobj.data[0][0].reshape(
                        lindblad_shape)
                oe_arrays.append(lindblad_op)
            oe_res = tn_expr(*oe_arrays, backend='jax')

            return oe_res.reshape(np.prod(self.dims), np.prod(self.dims))



from functools import partial
import jax
import jax.numpy as jnp
from tqdm import tqdm


def pass_operands(operands):
    return operands


def build_tqdm(n: int, print_rate: int = None, fwd_ad: bool = False):
    """Build the tqdm progress bar with backward mode auto-diff support.
    Args:
        n (int): total number of steps
        print_rate (int): the integer rate at which the progress bar will be updated
            by default the print rate will be 1/20 of the total number of steps
        fwd_ad (bool): whether to use forward mode auto-diff, default to False
    Returns:
        Callable: a wrapping function for progress bar
    """
    tqdm_bars = {}

    if print_rate is None:
        print_rate = n // 20 if n > 20 else 1
    else:
        if print_rate < 1:
            raise ValueError("print_rate should be a positive integer")
        print_rate = max(n, print_rate)

    remainder = n % print_rate

    def _id_tap_tqdm_progress(iter_num, message):

        def _define_tqdm(operand):
            if jax.process_index() == 0:
                tqdm_bars[0] = tqdm(range(n))
                tqdm_bars[0].set_description(message, refresh=False)

        def _update_tqdm(arg, operand):
            if jax.process_index() == 0:
                tqdm_bars[0].update(int(arg))

        _ = jax.lax.cond(
            iter_num == 0,
            partial(jax.debug.callback, _define_tqdm),
            pass_operands,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm by print rate
            (iter_num % print_rate == 0) & (iter_num != n - remainder),
            partial(jax.debug.callback, _update_tqdm, print_rate),
            pass_operands,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm by remainder
            iter_num == n - remainder,
            partial(jax.debug.callback, _update_tqdm, remainder),
            pass_operands,
            operand=None,
        )

    def _close_tqdm(operand):
        if tqdm_bars and jax.process_index() == 0:
            tqdm_bars[0].close()

    if fwd_ad:

        def _update_progress_bar(iter_num, result):
            """Updates tqdm in a JAX scan or loop."""
            message = f'Time evolution for {n:,} steps'
            _id_tap_tqdm_progress(iter_num, message)
            _ = jax.lax.cond(
                iter_num == n - 1,
                partial(jax.debug.callback, _close_tqdm),
                pass_operands,
                operand=None,
            )
            return iter_num, result
    else:

        @jax.custom_vjp
        def _update_progress_bar(iter_num, result):
            """Updates tqdm in a JAX scan or loop."""
            message = f'Forward time evolution for {n:,} steps'
            _id_tap_tqdm_progress(iter_num, message)
            _ = jax.lax.cond(
                iter_num == n - 1,
                partial(jax.debug.callback, _close_tqdm),
                pass_operands,
                operand=None,
            )
            return iter_num, result

        def _update_progress_bar_fwd(iter_num, result):
            return _update_progress_bar(iter_num, result), iter_num

        def _update_progress_bar_bwd(residual, cot_result):
            message = f'Backward time evolution for {n:,} steps'
            iter_num = n - residual - 1
            _id_tap_tqdm_progress(iter_num, message)
            _ = jax.lax.cond(
                iter_num == n - 1,
                partial(jax.debug.callback, _close_tqdm),
                pass_operands,
                operand=None,
            )
            return cot_result

        _update_progress_bar.defvjp(_update_progress_bar_fwd,
                                    _update_progress_bar_bwd)

    return _update_progress_bar


def scan_tqdm(func, xs, print_rate: int = None, fwd_ad: bool = False):
    """tqdm progress bar for a JAX scan.
    Args:
        func: function to be wrapped
        xs: scan steps
        print_rate (int): the integer rate at which the progress bar will be updated
            by default the print rate will be 1/20 of the total number of steps
        fwd_ad (bool): whether to use forward mode auto-diff, default to False
    Returns:
        Callable: a wrapping function for progress bar
        enumerate_xs: tuple(indices, xs)
    """
    n = len(xs)
    _update_progress_bar = build_tqdm(n, print_rate, fwd_ad)

    def wrapper_progress_bar(carry, enumerate_xs):
        iter_num, x = enumerate_xs
        result = func(carry, x)
        _, result = _update_progress_bar(iter_num, result)
        return result

    return wrapper_progress_bar, (jnp.arange(n), xs)



import os
import numpy as np
import jax
import haiku as hk

import supergrad
from supergrad.quantum_system import Fluxonium, Transmon

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)


def test_transmon_phase_basis():
    # Get cached data
    data_0 = np.load('data/data_create_transmon.npy')

    # construct model
    class ExploreTransmon(supergrad.Helper):

        def _init_quantum_system(self):
            self.tmon = Transmon(basis='phase',
                                 phiext=0.,
                                 truncated_dim=6,
                                 drive_for_state_phase='charge',
                                 n_max=31)

        def energy_spectrum(self):
            return self.tmon.eigenenergies()

    tmon = ExploreTransmon()
    # create params template
    ej_list = np.linspace(28, 32, 3)
    ec_list = np.linspace(1.0, 1.4, 3)
    ng_list = np.linspace(0.2, 0.4, 3)
    params = {'transmon': {'ec': ec_list, 'ej': ej_list, 'ng': ng_list}}

    multi_spec = jax.vmap(
        jax.vmap(
            jax.vmap(tmon.energy_spectrum,
                     in_axes=({
                         'transmon': {
                             'ec': None,
                             'ej': None,
                             'ng': 0
                         },
                     },)), ({
                         'transmon': {
                             'ec': 0,
                             'ej': None,
                             'ng': None
                         },
                     },)), ({
                         'transmon': {
                             'ec': None,
                             'ej': 0,
                             'ng': None
                         },
                     },))
    data_1 = multi_spec(params)

    assert np.allclose(data_0.flatten(), data_1.flatten())


def test_tunable_transmon_phase_basis():
    # Get cached data
    data_0 = np.load('data/data_create_tunable_transmon.npy')

    # construct model
    class ExploreTransmon(supergrad.Helper):

        def _init_quantum_system(self):
            self.tmon = Transmon(basis='phase',
                                 truncated_dim=6,
                                 d=hk.get_parameter('d', [], init=np.zeros),
                                 drive_for_state_phase='charge',
                                 n_max=40)

        def energy_spectrum(self):
            return self.tmon.eigenenergies()

    tmon = ExploreTransmon()
    # create params template
    ej_list = np.linspace(38, 42, 3)
    ec_list = np.linspace(0.1, 0.3, 3)
    d_list = np.linspace(0.05, 0.2, 3)
    flux_list = np.linspace(0.1, 0.5, 3)
    ng_list = np.linspace(0.2, 0.4, 3)
    params = {
        '~': {
            'd': d_list
        },
        'transmon': {
            'ec': ec_list,
            'ej': ej_list,
            'ng': ng_list,
            'phiext': flux_list * 2 * np.pi
        }
    }

    multi_spec = jax.vmap(
        jax.vmap(
            jax.vmap(
                jax.vmap(
                    jax.vmap(tmon.energy_spectrum,
                             in_axes=({
                                 '~': {
                                     'd': None,
                                 },
                                 'transmon': {
                                     'ec': None,
                                     'ej': None,
                                     'ng': 0,
                                     'phiext': None
                                 }
                             },)), ({
                                 '~': {
                                     'd': None,
                                 },
                                 'transmon': {
                                     'ec': None,
                                     'ej': None,
                                     'ng': None,
                                     'phiext': 0
                                 }
                             },)), ({
                                 '~': {
                                     'd': 0,
                                 },
                                 'transmon': {
                                     'ec': None,
                                     'ej': None,
                                     'ng': None,
                                     'phiext': None
                                 }
                             },)), ({
                                 '~': {
                                     'd': None,
                                 },
                                 'transmon': {
                                     'ec': 0,
                                     'ej': None,
                                     'ng': None,
                                     'phiext': None
                                 }
                             },)), ({
                                 '~': {
                                     'd': None,
                                 },
                                 'transmon': {
                                     'ec': None,
                                     'ej': 0,
                                     'ng': None,
                                     'phiext': None
                                 }
                             },))
    data_1 = multi_spec(params)

    assert np.allclose(data_0.flatten(), data_1.flatten())


def test_fluxonium_phase_basis():
    # Get cached data
    data_0 = np.load('data/data_create_fluxonium.npy')

    # construct model
    class ExploreFluxonium(supergrad.Helper):

        def _init_quantum_system(self):
            self.fmon = Fluxonium(basis='phase',
                                  truncated_dim=6,
                                  drive_for_state_phase='charge',
                                  put_phiext_on_inductor=False,
                                  phi_max=5 * np.pi)

        def energy_spectrum(self):
            return self.fmon.eigenenergies()

    tmon = ExploreFluxonium()
    # create params template
    ej_list = np.linspace(7, 10, 3)
    ec_list = np.linspace(2.2, 2.7, 3)
    el_list = np.linspace(0.3, 0.7, 3)
    flux_list = np.linspace(0.1, 0.5, 3)
    params = {
        'fluxonium': {
            'ec': ec_list,
            'ej': ej_list,
            'el': el_list,
            'phiext': flux_list * 2 * np.pi
        }
    }

    multi_spec = jax.vmap(
        jax.vmap(
            jax.vmap(
                jax.vmap(tmon.energy_spectrum,
                         in_axes=({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': None,
                                 'el': None,
                                 'phiext': 0,
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': None,
                                 'el': 0,
                                 'phiext': None
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': 0,
                                 'ej': None,
                                 'el': None,
                                 'phiext': None
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': 0,
                                 'el': None,
                                 'phiext': None
                             }
                         },))
    data_1 = multi_spec(params)
    assert np.allclose(data_0.flatten(), data_1.flatten())


def test_fluxonium_phase_basis_puton_inductor():
    # Get cached data
    data_0 = np.load('data/data_create_fluxonium.npy')

    # construct model
    class ExploreFluxonium(supergrad.Helper):

        def _init_quantum_system(self):
            self.fmon = Fluxonium(basis='phase',
                                  truncated_dim=6,
                                  drive_for_state_phase='charge',
                                  put_phiext_on_inductor=True,
                                  phi_max=5 * np.pi)

        def energy_spectrum(self):
            return self.fmon.eigenenergies()

    tmon = ExploreFluxonium()
    # create params template
    ej_list = np.linspace(7, 10, 3)
    ec_list = np.linspace(2.2, 2.7, 3)
    el_list = np.linspace(0.3, 0.7, 3)
    flux_list = np.linspace(0.1, 0.5, 3)
    params = {
        'fluxonium': {
            'ec': ec_list,
            'ej': ej_list,
            'el': el_list,
            'phiext': flux_list * 2 * np.pi
        }
    }

    multi_spec = jax.vmap(
        jax.vmap(
            jax.vmap(
                jax.vmap(tmon.energy_spectrum,
                         in_axes=({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': None,
                                 'el': None,
                                 'phiext': 0
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': None,
                                 'el': 0,
                                 'phiext': None
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': 0,
                                 'ej': None,
                                 'el': None,
                                 'phiext': None
                             }
                         },)), ({
                             'fluxonium': {
                                 'ec': None,
                                 'ej': 0,
                                 'el': None,
                                 'phiext': None
                             }
                         },))
    data_1 = multi_spec(params)
    assert np.allclose(data_0.flatten(), data_1.flatten(), atol=2e-3)



# %%
import os
import numpy as np
import jax.numpy as jnp
import haiku as hk
import pytest

import supergrad
from supergrad import basis
from supergrad.quantum_system import Fluxonium, parse_interaction, InteractingSystem
from supergrad.time_evolution.pulseshape import PulseCosine, PulseCosineRamping
from supergrad.time_evolution import sesolve, sesolve_final_states_w_basis_trans
from supergrad.utils.fidelity import compute_fidelity_with_1q_rotation_axis
from supergrad.utils import create_state_init, gates, identity_wrap

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)


def test_x_gate(solver='ode_expm', options={'astep': 1280}):

    class SingleGate(supergrad.Helper):

        def _init_quantum_system(self):
            self.fq = Fluxonium(phiext=0.5 * 2 * np.pi, phi_max=5 * np.pi)
            self.pulseshape = PulseCosine(modulate_wave=True)

        def evo(self):
            eig_val = self.fq.eigenenergies()
            ham_diag = jnp.diag(eig_val)

            opt = self.fq.phi_operator()
            ham = [ham_diag, [opt, self.pulseshape.create_pulse]]

            psi_list = jnp.array([basis(10, 0), basis(10, 1)])
            res = sesolve(ham,
                          psi_list, [0, self.pulseshape.length],
                          solver=solver,
                          options=options)
            return res

    def infidelity(params, all_params={}):
        params = hk.data_structures.merge(all_params, params)
        res = sg.evo(params)
        states = res[:, -1, :2, 0]
        fidelity, _ = compute_fidelity_with_1q_rotation_axis(gates.sigmax(),
                                                             states,
                                                             opt_method=None)
        return 1 - fidelity, states

    sg = SingleGate()

    params = {
        'fluxonium': {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(1. * 2 * jnp.pi)
        },
        'pulse_1mcos': {
            'length': jnp.array(40.),
            'amp': jnp.array(0.51753615),
            'omega_d': jnp.array(3.65911655),
            'phase': jnp.array(0.64509906)
        }
    }

    # load cached data
    u0 = np.load('data/data_x_gate.npy')
    infid0 = 1 - compute_fidelity_with_1q_rotation_axis(
        gates.sigmax(), u0, opt_method=None,
        compensation_option='arbit_single')[0]
    infid1, u1 = infidelity(params)
    relative_fid = compute_fidelity_with_1q_rotation_axis(u0,
                                                          u1,
                                                          opt_method=None)[0]
    assert np.allclose(infid0, infid1, atol=1e-4)
    assert 1 - relative_fid < 0.01


def test_x_gate_odeint():
    test_x_gate(solver='odeint', options={})


def test_cr_gate(solver='ode_expm',
                 options={
                     'astep': 2500,
                     'trotter_order': 1,
                     'diag_ops': False
                 }):

    class MultiGate(supergrad.Helper):

        def _init_quantum_system(self):
            self.control = Fluxonium(phiext=0.5 * 2 * np.pi,
                                     phi_max=5 * np.pi,
                                     name='control')
            self.target = Fluxonium(phiext=0.5 * 2 * np.pi,
                                    phi_max=5 * np.pi,
                                    name='target')
            inter_list = []

            inter_list.append(
                parse_interaction(op1=self.control.n_operator,
                                  op2=self.target.n_operator,
                                  name='jc'))
            inter_list.append(
                parse_interaction(op1=self.control.phi_operator,
                                  op2=self.target.phi_operator,
                                  name='jl'))

            self.hilbertspace = InteractingSystem([self.control, self.target],
                                                  inter_list)
            self.pulseshape = PulseCosineRamping(modulate_wave=True)

        def evo(self):
            # Single qubit product basis
            # construct static hamiltonian
            ham_static = self.hilbertspace.idling_hamiltonian_in_prod_basis()
            # add CR pulse
            opt = self.control.phi_operator
            wrap_opt = identity_wrap(opt, self.hilbertspace.subsystem_list)

            ham = [ham_static, [wrap_opt, self.pulseshape.create_pulse]]
            self.tlist = [0, self.pulseshape.pulse_endtime]
            u_to_eigen = self.hilbertspace.compute_transform_matrix()
            psi_list, _ = create_state_init(self._get_dims(), [[0, 2], [1, 2]])
            sim_u = sesolve_final_states_w_basis_trans(ham,
                                                       psi_list,
                                                       self.tlist,
                                                       u_to_eigen,
                                                       solver=solver,
                                                       options=options)
            return sim_u

    mg = MultiGate()
    params = {
        'control': {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(1.0 * 2 * jnp.pi),
        },
        'target': {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(0.8 * 2 * jnp.pi),
        },
        'jc': {
            'strength': jnp.array(11.5e-3 * 2 * jnp.pi)
        },
        'jl': {
            'strength': jnp.array(-1.0 * 2e-3 * 2 * jnp.pi)
        },
        'pulse_rampcos': {
            'amp': jnp.array(0.19362652),
            'omega_d': jnp.array(2.63342338),
            'phase': jnp.array(-0.56882493),
            't_plateau': jnp.array(69.9963988),
            't_ramp': jnp.array(30.20130596),
        }
    }

    def infidelity(params, all_params={}):
        params = hk.data_structures.merge(all_params, params)
        sim_u = mg.evo(params)
        fidelity, u_optimal = compute_fidelity_with_1q_rotation_axis(
            gates.cnot(),
            sim_u,
            opt_method='gd',
            compensation_option='arbit_single')
        return 1 - fidelity, sim_u

    # load cached data
    u0 = np.load('data/data_cnot_gate.npy')
    infid0 = 1 - compute_fidelity_with_1q_rotation_axis(
        gates.cnot(), u0, opt_method='gd',
        compensation_option='arbit_single')[0]
    infid1, u1 = infidelity(params)
    relative_fid = compute_fidelity_with_1q_rotation_axis(u0,
                                                          u1,
                                                          opt_method=None)[0]
    assert np.allclose(infid0, infid1, atol=1e-4)
    assert 1 - relative_fid < 0.01


@pytest.mark.slow
def test_cr_gate_odeint():
    test_cr_gate(solver='odeint', options={})


def test_cr_gate_diag():
    test_cr_gate(solver='ode_expm',
                 options={
                     'astep': 2500,
                     'trotter_order': 1,
                     'diag_ops': True
                 })


def test_cr_gate_trotter_2():
    test_cr_gate(solver='ode_expm', options={'astep': 2500, 'trotter_order': 2})


def test_cr_gate_trotter_4():
    test_cr_gate(solver='ode_expm', options={'astep': 2500, 'trotter_order': 4})


def test_cr_gate_trotter_4j():
    test_cr_gate(solver='ode_expm',
                 options={
                     'astep': 2500,
                     'trotter_order': 4j
                 })


@pytest.mark.slow
def test_cr_gate_no_trotter():
    test_cr_gate(solver='ode_expm',
                 options={
                     'astep': 2500,
                     'trotter_order': None
                 })
